#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE.txt file in the root directory of this source tree.

import random
import logging
from abc import ABC, abstractmethod
from contextlib import ExitStack
from pathlib import Path
import gzip
from typing import Any, Counter, Dict, Iterable, List, Optional, Tuple

import torch
from torchbiggraph.config import EntitySchema, RelationSchema
from torchbiggraph.converters.dictionary import Dictionary
from torchbiggraph.edgelist import EdgeList
from torchbiggraph.entitylist import EntityList
from torchbiggraph.graph_storages import (
    EDGE_STORAGES,
    ENTITY_STORAGES,
    RELATION_TYPE_STORAGES,
    AbstractEdgeAppender,
    AbstractEdgeStorage,
    AbstractEntityStorage,
    AbstractRelationTypeStorage,
)
# from torchbiggraph.types import UNPARTITIONED


class EdgelistReader(ABC):
    @abstractmethod
    def read(self, path: Path) -> Iterable[Tuple[str, str, Optional[str]]]:
        """Read rows from a path. Returns (lhs, rhs, rel)."""
        pass


class TSVEdgelistReader(EdgelistReader):
    def __init__(self, lhs_col: int, rhs_col: int, rel_col: int):
        self.lhs_col, self.rhs_col, self.rel_col = lhs_col, rhs_col, rel_col

    def read(self, path: Path):
        if str(path).endswith('tsv'):
            with path.open("rt") as tf:
                for line_num, line in enumerate(tf, start=1):
                    words = line.split()
                    # here is used for eliminating header line
                    if 'id' in words or 'node1' in words or 'node2' in words:
                        continue
                    try:
                        lhs_word = words[self.lhs_col]
                        rhs_word = words[self.rhs_col]
                        rel_word = words[self.rel_col] if self.rel_col is not None else None
                        yield lhs_word, rhs_word, rel_word
                    except IndexError:
                        raise RuntimeError(
                            f"Line {line_num} of {path} has only {len(words)} words"
                        ) from None
        elif str(path).endswith('gz'): # tsv.gz
            with gzip.open(str(path),'rt') as tf:
                for line_num, line in enumerate(tf, start=1):
                    words = line.split()
                    # here is used for eliminating header line
                    if 'id' in words or 'node1' in words or 'node2' in words:
                        continue
                    try:
                        lhs_word = words[self.lhs_col]
                        rhs_word = words[self.rhs_col]
                        rel_word = words[self.rel_col] if self.rel_col is not None else None
                        yield lhs_word, rhs_word, rel_word
                    except IndexError:
                        raise RuntimeError(
                            f"Line {line_num} of {path} has only {len(words)} words"
                        ) from None
        else:
            raise Exception("Invalid file!")


class ParquetEdgelistReader(EdgelistReader):
    def __init__(self, lhs_col: str, rhs_col: str, rel_col: Optional[str]):
        """Reads edgelists from a Parquet file.

        col arguments can either be the column name or the offset of the col.
        """
        self.lhs_col, self.rhs_col, self.rel_col = lhs_col, rhs_col, rel_col

    def read(self, path: Path):
        try:
            import parquet
        except ImportError as e:
            raise ImportError(
                f"{e}. HINT: You can install Parquet by running "
                "'pip install parquet'"
            )

        with path.open("rt") as tf:
            columns = [self.lhs_col, self.rhs_col]
            if self.rel_col is not None:
                columns.append(self.rel_col)
            for row in parquet.reader(tf, columns=columns):
                if self.rel_col is not None:
                    yield row
                else:
                    yield (row[0], row[1], None)


def collect_relation_types(
    relation_configs: List[RelationSchema],
    edge_paths: List[Path],
    dynamic_relations: bool,
    edgelist_reader: EdgelistReader,
    relation_type_min_count: int,
) -> Dictionary:

    if dynamic_relations:
        logging.info("Looking up relation types in the edge files...")
        counter: Counter[str] = Counter()
        for edgepath in edge_paths:
            for _lhs_word, _rhs_word, rel_word in edgelist_reader.read(edgepath):
                if rel_word is None:
                    raise RuntimeError("Need to specify rel_col in dynamic mode.")
                counter[rel_word] += 1

        logging.info(f"- Found {len(counter)} relation types")
        if relation_type_min_count > 0:
            logging.info(
                "- Removing the ones with fewer than "
                f"{relation_type_min_count} occurrences..."
            )
            counter = Counter(
                {k: c for k, c in counter.items() if c >= relation_type_min_count}
            )
            logging.info(f"- Left with {len(counter)} relation types")
        logging.info("- Shuffling them...")
        names = list(counter.keys())
        random.shuffle(names)

    else:
        names = [rconfig.name for rconfig in relation_configs]
        logging.info(f"Using the {len(names)} relation types given in the config")

    return Dictionary(names)


def collect_entities_by_type(
    relation_types: Dictionary,
    entity_configs: Dict[str, EntitySchema],
    relation_configs: List[RelationSchema],
    edge_paths: List[Path],
    dynamic_relations: bool,
    edgelist_reader: EdgelistReader,
    entity_min_count: int,
) -> Dict[str, Dictionary]:

    counters: Dict[str, Counter[str]] = {}
    for entity_name in entity_configs.keys():
        counters[entity_name] = Counter()

    logging.info("Searching for the entities in the edge files...")
    for edgepath in edge_paths:
        for lhs_word, rhs_word, rel_word in edgelist_reader.read(edgepath):
            if dynamic_relations or rel_word is None:
                rel_id = 0
            else:
                try:
                    rel_id = relation_types.get_id(rel_word)
                except KeyError:
                    raise RuntimeError("Could not find relation type in config")

            counters[relation_configs[rel_id].lhs][lhs_word] += 1
            counters[relation_configs[rel_id].rhs][rhs_word] += 1

    entities_by_type: Dict[str, Dictionary] = {}
    for entity_name, counter in counters.items():
        logging.info(f"Entity type {entity_name}:")
        logging.info(f"- Found {len(counter)} entities")
        if entity_min_count > 0:
            logging.info(
                f"- Removing the ones with fewer than {entity_min_count} occurrences..."
            )
            counter = Counter(
                {k: c for k, c in counter.items() if c >= entity_min_count}
            )
            logging.info(f"- Left with {len(counter)} entities")
        logging.info("- Shuffling them...")
        names = list(counter.keys())
        random.shuffle(names)
        entities_by_type[entity_name] = Dictionary(
            names, num_parts=entity_configs[entity_name].num_partitions
        )

    return entities_by_type


def generate_entity_path_files(
    entity_storage: AbstractEntityStorage,
    entities_by_type: Dict[str, Dictionary],
    relation_type_storage: AbstractRelationTypeStorage,
    relation_types: Dictionary,
    dynamic_relations: bool,
) -> None:
    logging.info(f"Preparing counts and dictionaries for entities and relation types:")
    entity_storage.prepare()
    relation_type_storage.prepare()

    for entity_name, entities in entities_by_type.items():
        for part in range(entities.num_parts):
            logging.info(
                f"- Writing count of entity type {entity_name} " f"and partition {part}"
            )
            entity_storage.save_count(entity_name, part, entities.part_size(part))
            entity_storage.save_names(entity_name, part, entities.get_part_list(part))

    if dynamic_relations:
        logging.info("- Writing count of dynamic relations")
        relation_type_storage.save_count(relation_types.size())
        relation_type_storage.save_names(relation_types.get_list())


def generate_edge_path_files(
    edge_file_in: Path,
    edge_path_out: Path,
    edge_storage: AbstractEdgeStorage,
    entities_by_type: Dict[str, Dictionary],
    relation_types: Dictionary,
    relation_configs: List[RelationSchema],
    dynamic_relations: bool,
    edgelist_reader: EdgelistReader,
) -> None:
    logging.info(
        f"Preparing edge path {edge_path_out}, "
        f"out of the edges found in {edge_file_in}"
    )
    edge_storage.prepare()

    num_lhs_parts = max(
        entities_by_type[rconfig.lhs].num_parts for rconfig in relation_configs
    )
    num_rhs_parts = max(
        entities_by_type[rconfig.rhs].num_parts for rconfig in relation_configs
    )

    logging.info(f"- Edges will be partitioned in {num_lhs_parts} x {num_rhs_parts} buckets.")

    processed = 0
    skipped = 0

    # We use an ExitStack in order to close the dynamically-created edge appenders.
    with ExitStack() as appender_stack:
        appenders: Dict[Tuple[int, int], AbstractEdgeAppender] = {}
        for lhs_word, rhs_word, rel_word in edgelist_reader.read(edge_file_in):
            if rel_word is None:
                rel_id = 0
            else:
                try:
                    rel_id = relation_types.get_id(rel_word)
                except KeyError:
                    # Ignore edges whose relation type is not known.
                    skipped += 1
                    continue

            if dynamic_relations:
                lhs_type = relation_configs[0].lhs
                rhs_type = relation_configs[0].rhs
            else:
                lhs_type = relation_configs[rel_id].lhs
                rhs_type = relation_configs[rel_id].rhs

            try:
                lhs_part, lhs_offset = entities_by_type[lhs_type].get_partition(
                    lhs_word
                )
                rhs_part, rhs_offset = entities_by_type[rhs_type].get_partition(
                    rhs_word
                )
            except KeyError:
                # Ignore edges whose entities are not known.
                skipped += 1
                continue

            if (lhs_part, rhs_part) not in appenders:
                appenders[lhs_part, rhs_part] = appender_stack.enter_context(
                    edge_storage.save_edges_by_appending(lhs_part, rhs_part)
                )
            appenders[lhs_part, rhs_part].append_edges(
                EdgeList(
                    EntityList.from_tensor(
                        torch.tensor([lhs_offset], dtype=torch.long)
                    ),
                    EntityList.from_tensor(
                        torch.tensor([rhs_offset], dtype=torch.long)
                    ),
                    torch.tensor([rel_id], dtype=torch.long),
                )
            )

            processed = processed + 1
            if processed % 100000 == 0:
                logging.info(f"- Processed {processed} edges so far...")

    logging.info(f"- Processed {processed} edges in total")
    if skipped > 0:
        logging.info(
            f"- Skipped {skipped} edges because their relation type or "
            f"entities were unknown (either not given in the config or "
            f"filtered out as too rare)."
        )


def convert_input_data(
    entity_configs: Dict[str, EntitySchema],
    relation_configs: List[RelationSchema],
    entity_path: str,
    edge_paths_out: List[str],
    edge_paths_in: List[Path],
    edgelist_reader: EdgelistReader,
    entity_min_count: int = 1,
    relation_type_min_count: int = 1,
    dynamic_relations: bool = False,
) -> None:
    if len(edge_paths_in) != len(edge_paths_out):
        raise ValueError(
            f"The edge paths passed as inputs ({edge_paths_in}) don't match "
            f"the ones specified as outputs ({edge_paths_out})"
        )

    entity_storage = ENTITY_STORAGES.make_instance(entity_path)
    relation_type_storage = RELATION_TYPE_STORAGES.make_instance(entity_path)
    edge_storages = [EDGE_STORAGES.make_instance(ep) for ep in edge_paths_out]

    some_files_exists = []
    some_files_exists.extend(
        entity_storage.has_count(entity_name, partition)
        for entity_name, entity_config in entity_configs.items()
        for partition in range(entity_config.num_partitions)
    )
    some_files_exists.extend(
        entity_storage.has_names(entity_name, partition)
        for entity_name, entity_config in entity_configs.items()
        for partition in range(entity_config.num_partitions)
    )
    if dynamic_relations:
        some_files_exists.append(relation_type_storage.has_count())
        some_files_exists.append(relation_type_storage.has_names())
    some_files_exists.extend(
        edge_storage.has_edges(0, 0)
        for edge_storage in edge_storages
    )

    if all(some_files_exists):
        logging.info(
            "Found some files that indicate that the input data "
            "has already been preprocessed, not doing it again."
        )
        all_paths = ", ".join(str(p) for p in [entity_path] + edge_paths_out)
        logging.info(f"These files are in: {all_paths}")
        return

    relation_types = collect_relation_types(
        relation_configs,
        edge_paths_in,
        dynamic_relations,
        edgelist_reader,
        relation_type_min_count,
    )

    entities_by_type = collect_entities_by_type(
        relation_types,
        entity_configs,
        relation_configs,
        edge_paths_in,
        dynamic_relations,
        edgelist_reader,
        entity_min_count,
    )

    generate_entity_path_files(
        entity_storage,
        entities_by_type,
        relation_type_storage,
        relation_types,
        dynamic_relations,
    )

    for edge_path_in, edge_path_out, edge_storage in zip(
        edge_paths_in, edge_paths_out, edge_storages
    ):
        generate_edge_path_files(
            edge_path_in,
            edge_path_out,
            edge_storage,
            entities_by_type,
            relation_types,
            relation_configs,
            dynamic_relations,
            edgelist_reader,
        )


def parse_config_partial(
    config_dict: Any,
) -> Tuple[Dict[str, EntitySchema], List[RelationSchema], str, List[str], bool]:
    entities_config = config_dict.get("entities")
    relations_config = config_dict.get("relations")
    entity_path = config_dict.get("entity_path")
    edge_paths = config_dict.get("edge_paths")
    dynamic_relations = config_dict.get("dynamic_relations", False)
    if not isinstance(entities_config, dict):
        raise TypeError("Config entities is not of type dict")
    if any(not isinstance(k, str) for k in entities_config.keys()):
        raise TypeError("Config entities has some keys that are not of type str")
    if not isinstance(relations_config, list):
        raise TypeError("Config relations is not of type list")
    if not isinstance(entity_path, str):
        raise TypeError("Config entity_path is not of type str")
    if not isinstance(edge_paths, list):
        raise TypeError("Config edge_paths is not of type list")
    if any(not isinstance(p, str) for p in edge_paths):
        raise TypeError("Config edge_paths has some items that are not of type str")
    if not isinstance(dynamic_relations, bool):
        raise TypeError("Config dynamic_relations is not of type bool")

    entities: Dict[str, EntitySchema] = {}
    relations: List[RelationSchema] = []
    for entity, entity_config in entities_config.items():
        entities[entity] = EntitySchema.from_dict(entity_config)
    for relation in relations_config:
        relations.append(RelationSchema.from_dict(relation))

    return entities, relations, entity_path, edge_paths, dynamic_relations


"""
Import Visual Genome into KGTK.
"""

import sys
from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Import Visual Genome into KGTK.' 
    }

def add_arguments(parser: KGTKArgumentParser):
    """
    Parse arguments
    Args:
            parser (argparse.ArgumentParser)
    """
    parser.add_input_file(positional=True, who="Visual Genome scene graph file")
    parser.add_input_file(who="Visual Genome file with attribute synsets.",
                            options=["--attr-synsets"], dest="attr_syn_file", metavar="ATTR_SYN_FILE")
    parser.add_output_file()

def run(input_file: KGTKFiles,
        attr_syn_file: KGTKFiles,
        output_file: KGTKFiles):

    # import modules locally
    import sys # type: ignore
    from kgtk.exceptions import kgtk_exception_auto_handler
    import csv
    import json
    import re
    from pathlib import Path
    from collections import defaultdict
    from kgtk.kgtkformat import KgtkFormat
    from kgtk.io.kgtkwriter import KgtkWriter

    def create_edge(node1, node1_lbl, node2, node2_lbl, rel, rel_lbl, image_id):
        my_row=[node1, rel, node2, '|'.join(node1_lbl), '|'.join(node2_lbl), rel_lbl, '', KgtkFormat.stringify('VG'), '']
        return my_row

    try:
        scene_graph_filename: Path = KGTKArgumentParser.get_input_file(input_file)
        attr_synsets_filename: Path = KGTKArgumentParser.get_input_file(attr_syn_file)

        out_columns=['node1', 'relation', 'node2', 'node1;label', 'node2;label','relation;label', 'relation;dimension', 'source', 'sentence']

        output_kgtk_file: Path = KGTKArgumentParser.get_output_file(output_file)
        ew: KgtkWriter = KgtkWriter.open(out_columns,
                                         output_kgtk_file,
                                         #mode=input_kr.mode,
                                         require_all_columns=False,
                                         prohibit_extra_columns=True,
                                         fill_missing_columns=True,
                                         gzip_in_parallel=False,
                                         #verbose=self.verbose,
                                         #very_verbose=self.very_verbose
                                         )

        proximity_relation='/r/LocatedNear'
        property_relation='mw:MayHaveProperty'
        property_relation_label=KgtkFormat.stringify('may have property')
        capableof_relation='/r/CapableOf'
        capableof_relation_label=KgtkFormat.stringify('capable of')


        with open(scene_graph_filename, 'r') as f:
            images_data=json.load(f)

        with open(attr_synsets_filename, 'r') as f:
            attr_synsets=json.load(f)

        for counter, an_image in enumerate(images_data):

            image_id=str(an_image['image_id'])
            
            # OBJECTS
            objid2names=defaultdict(list)
            objid2syns={}
            rows=[]
            for o in an_image['objects']:
                obj_id=o['object_id']
                o_synset=o['synsets']
                objid2syns[obj_id]=o_synset
                for name in o['names']:
                    name=name.strip().lower().rstrip('.')
                    if not name: continue
                    objid2names[obj_id].append(KgtkFormat.stringify(name))

                # ATTRIBUTES
                if 'attributes' in o.keys():
                    for attr in o['attributes']:
                        attr=attr.lower()
                        if attr in attr_synsets:
                            asyn=attr_synsets[attr]
                            apos=asyn.split('.')[1]
                            if apos!='n':
                                if apos=='v': # verb
                                    for osyn in o_synset:
                                        if osyn!=asyn:
                                            edge_row=create_edge('wn:' + osyn, 
                                                    objid2names[obj_id], 
                                                    'wn:' + asyn, 
                                                    [KgtkFormat.stringify(attr)], 
                                                    capableof_relation, 
                                                    capableof_relation_label, 
                                                    image_id)
                                            if edge_row not in rows:
                                                rows.append(edge_row)
                                else: #adjective
                                    for osyn in o_synset:
                                        if osyn!=asyn:
                                            edge_row=create_edge('wn:' + osyn, 
                                                    objid2names[obj_id], 
                                                    'wn:' + asyn, 
                                                    [KgtkFormat.stringify(attr)], 
                                                    property_relation, 
                                                    property_relation_label, 
                                                    image_id)
                                            if edge_row not in rows:
                                                rows.append(edge_row)
                   
            # RELATIONS
            for rel in an_image['relationships']:
                #synsets=rel['synsets']
                relation_label=KgtkFormat.stringify(rel['predicate'].lower().strip().strip('.'))
                sub_id=rel['subject_id']
                sub_names=objid2names[sub_id]
                sub_syns=objid2syns[sub_id]
                obj_id=rel['object_id']
                obj_names=objid2names[obj_id]
                obj_syns=objid2syns[obj_id]

                for ssyn in sub_syns:
                    for osyn in obj_syns:
                        if osyn!=ssyn:
                            edge_row=create_edge('wn:' + ssyn, 
                                    sub_names, 
                                    'wn:' + osyn, 
                                    obj_names, 
                                    proximity_relation, 
                                    relation_label, 
                                    image_id)
                            if edge_row not in rows:
                                rows.append(edge_row)
            for a_row in rows:
                ew.write(a_row)

        # Clean up
        ew.close()

    except Exception as e:
            kgtk_exception_auto_handler(e)



"""
Generate graph embedding based on Pytorch BigGraph library  

"""

from argparse import Namespace
from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
from kgtk.io.kgtkwriter import KgtkWriter
from kgtk.value.kgtkvalueoptions import KgtkValueOptions
import attr
import typing
from kgtk.kgtkformat import KgtkFormat
from pathlib import Path
import sys
import logging
import os

# remove the Issue: Initializing libiomp5.dylib, but found libiomp5.dylib already initialized.
os.environ['KMP_DUPLICATE_LIB_OK'] = 'True'


@attr.s(slots=True, frozen=False)
class KgtkCreateTmpTsv(KgtkFormat):
    input_file_path: Path = attr.ib(validator=attr.validators.instance_of(Path))
    output_file_path: Path = attr.ib(validator=attr.validators.instance_of(Path))

    reader_options: typing.Optional[KgtkReaderOptions] = attr.ib(default=None)
    value_options: typing.Optional[KgtkValueOptions] = attr.ib(default=None)

    error_file: typing.TextIO = attr.ib(default=sys.stderr)
    verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
    very_verbose: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)

    def process(self):
        # Open the input file.
        if self.verbose:
            print("Opening the input file: %s" % str(self.input_file_path), file=self.error_file, flush=True)

        kr: KgtkReader = KgtkReader.open(self.input_file_path,
                                         error_file=self.error_file,
                                         options=self.reader_options,
                                         value_options=self.value_options,
                                         verbose=self.verbose,
                                         very_verbose=self.very_verbose,
                                         )

        if self.verbose:
            print("Opening the output file: %s" % str(self.output_file_path), file=self.error_file, flush=True)

        # Open the output file.
        kw: KgtkWriter = KgtkWriter.open(kr.column_names,
                                         self.output_file_path,
                                         mode=KgtkWriter.Mode[kr.mode.name],
                                         require_all_columns=False,
                                         prohibit_extra_columns=True,
                                         fill_missing_columns=False,
                                         gzip_in_parallel=False,
                                         verbose=self.verbose,
                                         very_verbose=self.very_verbose)
        # here kw has one line already where PBG doesn't need it,

        input_line_count: int = 0
        if self.verbose:
            print("Processing the input records.", file=self.error_file, flush=True)

        # node1 relation node2
        node1_index = kr.get_node1_column_index()
        node2_index = kr.get_node2_column_index()
        relation_index = kr.get_label_column_index()

        row: typing.List[str]

        for row in kr:
            input_line_count += 1
            kw.write([row[node1_index], row[relation_index], row[node2_index]])

        if self.verbose:
            print("Processed %d records." % input_line_count, file=self.error_file, flush=True)

        kw.close()


def get_config(**kwargs):
    """
    configurations for graph embedding
    """

    output_folder = kwargs['temporary_directory'] / 'output'
    entity_path = str(output_folder)
    edge_paths = [str((output_folder / 'edges_partitioned'))]
    checkpoint_path = str((output_folder / 'model'))

    config = dict(
        # I/O data
        entity_path=entity_path,
        edge_paths=edge_paths,
        checkpoint_path=checkpoint_path,
        # Graph structure
        entities={"all": {"num_partitions": 1}},
        relations=[  # relation template setting
            {
                "name": "all_edges",
                "lhs": "all",
                "rhs": "all",
                "operator": kwargs['operator'],
            }
        ],
        dynamic_relations=kwargs['dynamic_relaitons'],
        # Scoring model
        dimension=kwargs['dimension_num'],
        global_emb=kwargs['global_emb'],
        comparator=kwargs['comparator'],
        # Training
        init_scale=kwargs['init_scale'],
        bias=kwargs['bias'],
        num_epochs=kwargs['num_epochs'],
        loss_fn=kwargs['loss_fn'],
        lr=kwargs['learning_rate'],
        # Evaluation during training
        eval_fraction=kwargs['eval_fraction'],  # to reproduce results, we need to use all training data 
    )

    return config


def parser():
    return {
        'help': 'graph embedding functionality',
        'description': 'Generate graph embedding in kgtk tsv format, here we use PytorchBigGraph as \
        low-level implementation '
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    # import modules locally
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
    from kgtk.utils.argparsehelpers import optional_bool
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert

    # IO
    parser.add_input_file()
    parser.add_output_file()

    parser.add_argument('-l', "--log", dest="log_file_path",
                        help="Setting the log path [Default: None]",
                        type=Path, default=None, metavar="")
    parser.add_argument('-T', '--temporary_directory', dest='temporary_directory',
                        help="Sepecify the directory location to store temporary file",
                        type=Path, default=Path('/tmp/'), metavar='')
    parser.add_argument('-ot', '--output_format', dest='output_format',
                        help="Outputformat for embeddings [Default: w2v] Choice: kgtk | w2v | glove",
                        default='w2v', metavar='')
    parser.add_argument('-r', '--retain_temporary_data', dest='retain_temporary_data',
                        help="When opearte graph, some tempory files will be generated, set True to retain these files",
                        type=bool, default=True, metavar='True|False')
    # Training parameters
    parser.add_argument('-d', "--dimension", dest="dimension_num",
                        help="Dimension of the real space the embedding live in [Default: 100]",
                        type=int, default=100, metavar="")
    parser.add_argument('-s', "--init_scale", dest="init_scale",
                        help="Generating the initial embedding with this standard deviation [Default: 0.001]" +
                             "If no initial embeddings are provided, they are generated by sampling each dimension" +
                             "from a centered normal distribution having this standard deviation.",
                        type=float, default=0.001, metavar="")
    parser.add_argument('-c', '--comparator', dest='comparator',
                        help="How the embeddings of the two sides of an edge (after having already " +
                             "undergone some processing) are compared to each other to produce a score[Default: dot]," +
                             "Choice: dot|cos|l2|squared_l2",
                        default='dot', choices=['dot', 'cos', 'l2', 'squared_l2'], metavar='dot|cos|l2|squared_l2')
    parser.add_argument('-op', '--operator', dest='operator',
                        help="The transformation to apply to the embedding of one of the sides of the edge " +
                             "(typically the right-hand one) before comparing it with the other one. It reflects"
                             "which model that embedding uses. [Default:ComplEx]", default='ComplEx',
                        metavar='RESCAL|DistMult|ComplEx|TransE')
    parser.add_argument('-e', '--num_epochs', dest='num_epochs',
                        help="The number of times the training loop iterates over all the edges.[Default:100]",
                        type=int, default=100, metavar='')
    parser.add_argument('-b', '--bias', dest='bias',
                        help="Whether use the bias choice [Default: False],If enabled, withhold the first " +
                             "dimension of the embeddings from the comparator and instead use it as a bias, adding " +
                             "back to the score. Makes sense for logistic and softmax loss functions. ",
                        type=bool, default=False, metavar='True|False')
    parser.add_argument('-w', '--workers', dest='workers',
                        help="The number of worker processes for training. If not given, set to CPU count.",
                        type=int, default=None, metavar='')
    parser.add_argument('-bs', '--batch_size', dest='batch_size',
                        help="The number of edges per batch.[Default:1000]",
                        type=int, default=1000, metavar='')
    parser.add_argument('-lf', '--loss_fn', dest='loss_fn',
                        help="How the scores of positive edges and their corresponding negatives " +
                             "are evaluated.[Default: ranking], Choice: ranking|logistic|softmax",
                        # default will be setting to ranking later
                        default=None, choices=['ranking', 'logistic', 'softmax', None],
                        metavar='ranking|logistic|softmax')
    parser.add_argument('-lr', '--learning_rate', dest='learning_rate',
                        help="The learning rate for the optimizer.[Default: 0.1]",
                        # default will be setting to 0.1 later
                        type=float, default=None, metavar='')
    parser.add_argument('-ef', '--eval_fraction', dest='eval_fraction',
                        help="The fraction of edges withheld from training and used to track evaluation " +
                             "metrics during training. [Defalut:0.0 training all edges ]",
                        type=float, default=0.0, metavar='')
    parser.add_argument('-dr', '--dynamic_relaitons', dest='dynamic_relaitons',
                        help="Whether use dynamic relations (when graphs with a " +
                             "large number of relations) [Default: True]",
                        type=bool, default=True, metavar='True|False')
    parser.add_argument('-ge', '--global_emb', dest='global_emb',
                        help="Whether use global embedding, if enabled, add to each embedding a vector that is common "
                             "to all the entities of a certain type. This vector is learned "
                             "during training.[Default: False] ", type=bool, default=False, metavar='True|False')
    # kgtk format
    parser.add_argument("--no-output-header", dest="output_no_header", metavar="True|False",
                        help="When true, do not write a header to the output file (default=%(default)s).",
                        type=optional_bool, nargs='?', const=True, default=False)

    KgtkReader.add_debug_arguments(parser)
    KgtkReaderOptions.add_arguments(parser,
                                    mode_options=True,
                                    default_mode=KgtkReaderMode[parsed_shared_args._mode],
                                    expert=_expert)
    KgtkValueOptions.add_arguments(parser)


def config_preprocess(raw_config):
    # Setting learning rate and loss function for different algo if user doesn't assign them
    '''
    loss_fn: ranking
    learning_rate: 0.01
    operator:complex_diagonal
    '''

    algorithm_operator = {"complex": "complex_diagonal",
                          "distmult": "diagonal",
                          "rescal": "linear",
                          "transe": "translation"}
    try:
        algorithm = algorithm_operator[raw_config['relations'][0]['operator'].lower()]
        raw_config['relations'][0]['operator'] = algorithm
    except Exception:
        print('Plase use valid operator! choices: RESCAL|DistMult|ComplEx|TransE', file=sys.stderr, flush=True)
        sys.exit()

    loss_fn = raw_config['loss_fn']
    learning_rate = raw_config['lr']
    if algorithm and loss_fn and learning_rate:
        return

    if not algorithm:  # use doesn't enter anything
        raw_config['relations'][0]['operator'] = 'complex_diagonal'
        if not loss_fn:
            raw_config['loss_fn'] = 'logistic'
        if not learning_rate:
            raw_config['lr'] = 0.1
    if algorithm:  # give the algorithm
        if algorithm == 'complex_diagonal':  # ComplEx
            if not loss_fn:
                raw_config['loss_fn'] = 'logistic'
            if not learning_rate:
                raw_config['lr'] = 0.1
        elif algorithm == 'translation':  # TransE
            if not loss_fn:
                raw_config['loss_fn'] = 'logistic'
            if not learning_rate:
                raw_config['lr'] = 0.1
        elif algorithm == 'diagonal':  # DistMult
            if not loss_fn:
                raw_config['loss_fn'] = 'ranking'
            if not learning_rate:
                raw_config['lr'] = 0.01
        else:  # RESCAL
            if not loss_fn:
                raw_config['loss_fn'] = 'ranking'
            if not learning_rate:
                raw_config['lr'] = 0.01

    processed_config = raw_config
    return processed_config


# convert wv format to kgtk format ..
def generate_kgtk_output(entities_output, output_kgtk_file, output_no_header, verbose, very_verbose, error_file):
    # Open the output file.
    kw: KgtkWriter = KgtkWriter.open(
        ['node1', 'label', 'node2'],
        output_kgtk_file,
        mode=KgtkWriter.Mode.AUTO,
        require_all_columns=False,
        prohibit_extra_columns=False,
        fill_missing_columns=False,
        gzip_in_parallel=False,
        no_header=output_no_header,
        verbose=verbose,
        very_verbose=very_verbose)

    input_line_count: int = 0
    if verbose:
        logging.info("Processing the input records.", file=error_file, flush=True)

    MODULE_NAME = 'graph_embeddings'  # __name__.split('.')[-1]
    with open(entities_output) as wv_file:
        for line in wv_file:
            vals = line.strip().split('\t')
            entity_name = vals[0]
            entity_vev = ','.join(vals[1:])
            input_line_count += 1
            kw.write([entity_name, MODULE_NAME, entity_vev])

    if verbose:
        logging.info("Processed %d records." % (input_line_count), file=error_file, flush=True)

    kw.close()


def generate_w2v_output(entities_output, output_kgtk_file, kwargs):
    fout = open(output_kgtk_file, 'w')
    fin = open(entities_output)
    entity_num = len(fin.readlines())
    fin.close()
    fout.write(str(entity_num) + ' ' + str(kwargs['dimension_num']) + '\n')
    with open(entities_output) as fin:
        for line in fin:
            embedding = ' '.join(line.split('\t'))
            fout.write(embedding)
    fout.close()


def run(input_file: KGTKFiles,
        output_file: KGTKFiles,

        verbose: bool = False,
        very_verbose: bool = False,
        **kwargs):
    """
    **kwargs stores all parameters providing by user
    """
    # print(kwargs)

    # import modules locally
    import sys
    import typing
    import logging
    from pathlib import Path
    import os
    import shutil
    from torchbiggraph.config import parse_config
    from kgtk.exceptions import KGTKException
    # copy  missing file under kgtk/graph_embeddings
    from kgtk.graph_embeddings.importers import TSVEdgelistReader, convert_input_data
    from torchbiggraph.train import train
    from torchbiggraph.util import SubprocessInitializer, setup_logging
    from kgtk.graph_embeddings.export_to_tsv import make_tsv
    # from torchbiggraph.converters.export_to_tsv import make_tsv

    try:
        input_kgtk_file: Path = KGTKArgumentParser.get_input_file(input_file)
        output_kgtk_file: Path = KGTKArgumentParser.get_output_file(output_file)

        # store the data into log file, then the console will not output anything
        if kwargs['log_file_path'] is not None:
            log_file_path = kwargs['log_file_path']
            logging.basicConfig(format='%(asctime)s - %(filename)s[line:%(lineno)d] \
            - %(levelname)s: %(message)s',
                                level=logging.DEBUG,
                                filename=str(log_file_path),
                                filemode='w')
            print(f'In Processing, Please go to {kwargs["log_file_path"]} to check details', file=sys.stderr,
                  flush=True)

        tmp_folder = kwargs['temporary_directory']
        tmp_tsv_path: Path = tmp_folder / f'tmp_{input_kgtk_file.name}'
        # tmp_tsv_path:Path = input_kgtk_file.parent/f'tmp_{input_kgtk_file.name}'

        #  make sure the tmp folder exists, otherwise it will raise an exception
        if not os.path.exists(tmp_folder):
            os.makedirs(tmp_folder)

        try:  # if output_kgtk_file is not empty, delete it
            output_kgtk_file.unlink()
        except Exception:
            pass  # didn't find, then let it go

        # *********************************************
        # 0. PREPARE PBG TSV FILE
        # *********************************************
        reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
        value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)
        error_file: typing.TextIO = sys.stdout if kwargs.get("errors_to_stdout") else sys.stderr
        kct: KgtkCreateTmpTsv = KgtkCreateTmpTsv(input_file_path=input_kgtk_file,
                                                 output_file_path=tmp_tsv_path,
                                                 reader_options=reader_options,
                                                 value_options=value_options,
                                                 error_file=error_file,
                                                 verbose=verbose,
                                                 very_verbose=very_verbose,
                                                 )
        # prepare the graph file
        # create a tmp tsv file for PBG embedding

        logging.info('Generate the valid tsv format for embedding ...')
        kct.process()
        logging.info('Embedding file is ready...')

        # *********************************************
        # 1. DEFINE CONFIG  
        # *********************************************
        raw_config = get_config(**kwargs)

        # setting corresponding learning rate and loss function for different algorthim
        processed_config = config_preprocess(raw_config)

        # temporry output folder 
        tmp_output_folder = Path(processed_config['entity_path'])

        # before moving, need to check whether the tmp folder is not empty in case of bug
        try:  # if temporry output folder is alrady existing then delete it
            shutil.rmtree(tmp_output_folder)
        except Exception:
            pass  # didn't find, then let it go

        # **************************************************
        # 2. TRANSFORM GRAPH TO A BIGGRAPH-FRIENDLY FORMAT
        # **************************************************
        setup_logging()
        config = parse_config(processed_config)
        subprocess_init = SubprocessInitializer()
        input_edge_paths = [tmp_tsv_path]

        convert_input_data(
            config.entities,
            config.relations,
            config.entity_path,
            config.edge_paths,
            input_edge_paths,
            TSVEdgelistReader(lhs_col=0, rel_col=1, rhs_col=2),
            dynamic_relations=config.dynamic_relations,
        )

        # ************************************************
        # 3. TRAIN THE EMBEDDINGS
        # *************************************************
        train(config, subprocess_init=subprocess_init)

        # ************************************************
        # 4. GENERATE THE OUTPUT
        # ************************************************
        # entities_output = output_kgtk_file
        entities_output = tmp_output_folder / 'entities_output.tsv'
        relation_types_output = tmp_output_folder / 'relation_types_tf.tsv'

        with open(entities_output, "xt") as entities_tf, open(
                relation_types_output, "xt"
        ) as relation_types_tf:
            make_tsv(config, entities_tf, relation_types_tf)

        # output  correct format for embeddings
        if kwargs['output_format'] == 'glove':  # glove format output
            shutil.copyfile(entities_output, output_kgtk_file)
        elif kwargs['output_format'] == 'w2v':  # w2v format output
            generate_w2v_output(entities_output, output_kgtk_file, kwargs)

        else:  # write to the kgtk output format tsv
            generate_kgtk_output(entities_output,
                                 output_kgtk_file,
                                 kwargs.get('output_no_header', False),
                                 verbose,
                                 very_verbose,
                                 kct.error_file)

        logging.info(f'Embeddings has been generated in {output_kgtk_file}.')

        # ************************************************
        # 5. Garbage collection  
        # ************************************************
        if kwargs['retain_temporary_data'] is False:
            shutil.rmtree(kwargs['temporary_directory'])
            # tmp_tsv_path.unlink() # delete temporay tsv file
            # shutil.rmtree(tmp_output_folder) # deleter temporay output folder   

        if kwargs["log_file_path"] is not None:
            print('Processed Finished.', file=sys.stderr, flush=True)
            logging.info(f"Process Finished.\nOutput has been saved in {repr(str(output_kgtk_file))}")
        else:
            print(f"Process Finished.\nOutput has been saved in {repr(str(output_kgtk_file))}", file=sys.stderr,
                  flush=True)

    except Exception as e:
        raise KGTKException(str(e))



from argparse import ArgumentParser, RawDescriptionHelpFormatter, Namespace, SUPPRESS, _StoreTrueAction
from functools import partial
from pathlib import Path
import typing

from kgtk.exceptions import KGTKArgumentParseException, KGTKSyntaxException, KGTKDependencyException


KGTKFiles = typing.Optional[typing.Union[Path,
                                         typing.List[typing.Union[typing.Optional[Path],
                                                                  typing.List[Path]]]]]

class KGTKArgumentParser(ArgumentParser):
    SUPPORT_POSITIONAL_ARGS: bool = False
    DEFAULT_INPUT_FILE_WHO: str = "KGTK input file"
    DEFAULT_OUTPUT_FILE_WHO: str = "KGTK output file"

    def __init__(self, *args, **kwargs):
        if not kwargs.get('formatter_class'):
            kwargs['formatter_class'] = RawDescriptionHelpFormatter

        self.shared_arguments = set()
        self.default_arguments = set()
        self.add_default_argument_funcs = {}

        super(KGTKArgumentParser, self).__init__(*args, **kwargs)

    def add_default_argument(self, *args, **kwargs):
        if 'dest' not in kwargs:
            raise KGTKSyntaxException('Default argument error: dest not defined')
        self.add_default_argument_funcs[kwargs['dest']] = partial(self.add_argument, *args, **kwargs)

    def accept_shared_argument(self, dest):
        # Shared arguments that have been accepted by the subcommand
        # will be passed as keyword arguments to the subcommand
        # in cli_entry.cli_entry_pipe(...).
        self.shared_arguments.add(dest)

    def accept_default_argument(self, dest):
        # load default arguments
        self.add_default_argument_funcs[dest]()

    def exit(self, status=0, message=None):
        if status == 2:
            status = KGTKArgumentParseException.return_code
        super(KGTKArgumentParser, self).exit(status, message)

    def add_file(self,
                 who: typing.Optional[str],
                 dest: str,
                 options: typing.List[str],
                 metavar: str,
                 optional: bool,
                 allow_list: bool,
                 positional: bool,
                 default_help: str,
                 stdio_name: str,
                 default_stdio: bool,
                 allow_stdio: bool,
    ):

        helpstr: str
        if who is None:
            helpstr = "The " + default_help + "."
        elif who == SUPPRESS:
            helpstr = SUPPRESS
        else:
            helpstr = who

        if optional:
            # Not required, no default. Positional output not allowed, lists not allowed.
            if allow_stdio:
                if helpstr != SUPPRESS:
                    helpstr += " (Optional, use '-' for %s.)" % stdio_name
                self.add_argument(*options, dest=dest, type=Path, metavar=metavar, help=helpstr)
            else:
                if helpstr != SUPPRESS:
                    helpstr += " (Optional)"
                self.add_argument(*options, dest=dest, type=Path, metavar=metavar, help=helpstr)
            return

        # This is a required file, defaulting to stdio.
        if helpstr != SUPPRESS:
            if allow_stdio:
                if default_stdio:
                    helpstr += " (May be omitted or '-' for %s.)" % stdio_name
                else:
                    helpstr += " (Required, use '-' for %s.)" % stdio_name
            else:
                helpstr += " (Required)"

        positional &= self.SUPPORT_POSITIONAL_ARGS
        if positional:
            helpstr2: str
            if helpstr == SUPPRESS:
                helpstr2 = SUPPRESS
            else:
                helpstr2 = helpstr + " (Deprecated, use %s %s)" % (options[0], metavar)            

        if allow_list:
            if positional:
                self.add_argument(dest, nargs="*", type=Path, help=helpstr2, metavar=metavar, action="append")
                self.add_argument(*options, dest=dest, nargs="+", type=Path, metavar=metavar, help=helpstr, action="append")
            else:
                if allow_stdio and default_stdio:
                    self.add_argument(*options, dest=dest, nargs="+", type=Path, metavar=metavar, help=helpstr, default=[], action="append")
                else:
                    self.add_argument(*options, dest=dest, nargs="+", type=Path, metavar=metavar, help=helpstr, required=True, action="append")

        else:
            if positional:
                self.add_argument(dest, nargs="?", type=Path, help=helpstr2, metavar=metavar, action="append")
                self.add_argument(*options, dest=dest, type=Path, metavar=metavar, help=helpstr, action="append")
            else:
                if allow_stdio and default_stdio:
                    self.add_argument(*options, dest=dest, type=Path, metavar=metavar, help=helpstr, default=Path("-"))
                else:
                    self.add_argument(*options, dest=dest, type=Path, metavar=metavar, help=helpstr, required=True)


    def add_input_file(self,
                       who: typing.Optional[str] = None,
                       dest: str = "input_file",
                       options: typing.List[str] = ["-i", "--input-file"],
                       metavar: str = "INPUT_FILE",
                       optional: bool = False,
                       allow_list: bool = False,
                       positional: bool = False,
                       default_stdin: bool = True,
                       allow_stdin: bool = True,
    ):

        return self.add_file(who, dest, options, metavar, optional, allow_list, positional, self.DEFAULT_INPUT_FILE_WHO, "stdin", default_stdin, allow_stdin)

    def add_output_file(self,
                        who: typing.Optional[str] = None,
                        dest: str = "output_file",
                        options: typing.List[str] = ["-o", "--output-file"],
                        metavar: str = "OUTPUT_FILE",
                        optional: bool = False,
                        allow_list: bool = False,
                        positional: bool = False,
                        default_stdout: bool = True,
                        allow_stdout: bool = True,
    ):
        return self.add_file(who, dest, options, metavar, optional, allow_list, positional, self.DEFAULT_OUTPUT_FILE_WHO, "stdout", default_stdout, allow_stdout)

    @classmethod
    def get_path_list(cls, paths: KGTKFiles, who: str, default_stdio: bool)->typing.List[Path]:
        # Builds a list of paths from the awkward possible returns from argument parsing.
        if paths is None:
            if default_stdio:
                return [ Path("-") ]
            else:
                return [ ]
        elif isinstance(paths, Path):
            return [ paths ]
        elif isinstance(paths, list):
            result: typing.List[Path] = [ ]
            pl: typing.Union[typing.Optional[Path], typing.List[Path]]
            for pl in paths:
                if pl is None:
                    continue
                if isinstance(pl, Path):
                    result.append(pl)
                elif isinstance(pl, list):
                    result.extend(pl)
                else:
                    raise KGTKArgumentParseException("%s: Unexpected component '%s' in path list '%s'." % (who, str(pl), str(paths)))
            if len(result) == 0 and default_stdio:
                return [ Path("-") ]
            else:
                return result
        else:
            raise KGTKArgumentParseException("%s: Unexpected path list '%s'." % (who, str(paths)))

    @classmethod
    def get_input_file(cls,
                       paths: KGTKFiles,
                       who: typing.Optional[str] = None,
                       default_stdin: bool = True,
                       
    )->Path:
    
        if who is None:
            who = cls.DEFAULT_INPUT_FILE_WHO

        p: typing.List[Path] = cls.get_path_list(paths, who, default_stdio=default_stdin)
        if len(p) == 1:
            return p[0]
        elif len(p) == 0:
            raise KGTKArgumentParseException("%s: Please supply a filename path." % who)
        else:
            raise KGTKArgumentParseException("%s: Too many files: '%s'" % (who, str(paths)))

                                
    @classmethod
    def get_optional_input_file(cls,
                                paths: KGTKFiles,
                                who: typing.Optional[str] = None,
                       
    )->typing.Optional[Path]:
    
        if who is None:
            who = cls.DEFAULT_INPUT_FILE_WHO

        p: typing.List[Path] = cls.get_path_list(paths, who, default_stdio=False)
        if len(p) == 0:
            return None                                
        elif len(p) == 1:
            return p[0]
        else:
            raise KGTKArgumentParseException("%s: Too many files: '%s'" % (who, str(paths)))

    @classmethod
    def get_input_file_list(cls,
                            paths: KGTKFiles,
                            who: typing.Optional[str] = None,
                            default_stdin: bool = True,
                       
    )->typing.List[Path]:
    
        if who is None:
            who = cls.DEFAULT_INPUT_FILE_WHO

        return cls.get_path_list(paths, who, default_stdio=default_stdin)


    @classmethod
    def get_output_file(cls,
                       paths: KGTKFiles,
                       who: typing.Optional[str] = None,
                       default_stdout: bool = True,
                       
    )->Path:
    
        if who is None:
            who = cls.DEFAULT_OUTPUT_FILE_WHO

        p: typing.List[Path] = cls.get_path_list(paths, who, default_stdio=default_stdout)
        if len(p) == 1:
            return p[0]
        elif len(p) == 0:
            raise KGTKArgumentParseException("%s: Please supply a filename path." % who)
        else:
            raise KGTKArgumentParseException("%s: Too many files: '%s'" % (who, str(paths)))

                                
    @classmethod
    def get_optional_output_file(cls,
                                paths: KGTKFiles,
                                who: typing.Optional[str] = None,
                       
    )->typing.Optional[Path]:
    
        if who is None:
            who = cls.DEFAULT_OUTPUT_FILE_WHO

        p: typing.List[Path] = cls.get_path_list(paths, who, default_stdio=False)
        if len(p) == 0:
            return None                                
        elif len(p) == 1:
            return p[0]
        else:
            raise KGTKArgumentParseException("%s: Too many files: '%s'" % (who, str(paths)))

    @classmethod
    def get_output_file_list(cls,
                            paths: KGTKFiles,
                            who: typing.Optional[str] = None,
                            default_stdout: bool = True,
                       
    )->typing.List[Path]:
    
        if who is None:
            who = cls.DEFAULT_OUTPUT_FILE_WHO

        return cls.get_path_list(paths, who, default_stdio=default_stdout)


class CheckDepsAction(_StoreTrueAction):

    def __call__(self, parser, namespace, values, option_string=None):
        from kgtk.dependencies import check_dependencies
        if check_dependencies():
            parser.exit()
        parser.exit(KGTKDependencyException.return_code)


def add_shared_arguments(parser):
    # set shared arguments here
    # 1. no flag, only name (starts with two dashes: --XXX)
    # 2. need to have default value
    # 3. need to specify dest, which value starts with underscore
    # e.g., parser.add_argument('--debug', dest='_debug', action='store_true', default=False, help='enable debug mode')
    pass


def add_default_arguments(parser):
    # set default arguments here
    # need to specify dest
    # e.g., parser.add_default_argument('--save', dest='save', action='store', help='save to file')
    pass



import re
from typing import Optional, Union
import rdflib.namespace
from rdflib.namespace import Namespace, OWL, RDF, XSD
from rdflib import URIRef
import warnings
from kgtk.exceptions import WrongFormatURIException, PrefixNotFoundException, PrefixAlreadyUsedException
from kgtk.exceptions import SplitURIWithUnknownPrefix
from kgtk.knowledge_graph.node import URI


SCHEMA = Namespace('http://schema.org/')
DIG = Namespace('http://dig.isi.edu/ontologies/dig/')

URI_PATTERN = re.compile(r'^http:|^urn:|^info:|^ftp:|^https:')
URI_ABBR_PATTERN = re.compile(r'^(?:([^:]*):)?([^:]+)$')


class NamespaceManager(rdflib.namespace.NamespaceManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.graph.namespace_manager = self

    def parse_uri(self, text: Union[str, URI]) -> URIRef:
        """
        Parse input text into URI

        :param text: can be one of
              1. URI, directly return
              2. prefix:name, query namespace for prefix, return expanded URI
              3. name, use default namespace to expand it and return it
        :return: URIRef
        """
        if self.check_uriref(text):
            return self.check_uriref(text)
        elif isinstance(text, str):
            text = text.strip()
            m = URI_ABBR_PATTERN.match(text)
            if m:
                prefix, name = m.groups()
                base = self.store.namespace(prefix if prefix else '')
                if not base:
                    raise PrefixNotFoundException("Prefix: %s", prefix)
                return URIRef(base + name)
        elif isinstance(text, URI):
            return self.parse_uri(text.value)
        raise WrongFormatURIException(text)

    def bind(self, prefix: str, namespace: str, override=True, replace=True):
        """
        bind a given namespace to the prefix, forbids same prefix with different namespace

        :param prefix:
        :param namespace:
        :param override: if override, rebind, even if the given namespace is already bound to another prefix.
        :param replace: if replace, replace any existing prefix with the new namespace
        """
        namespace = URIRef(str(namespace))
        # When documenting explain that override only applies in what cases
        if prefix is None:
            prefix = ''
        bound_namespace = self.store.namespace(prefix)
        # Check if the bound_namespace contains a URI and if so convert it into a URIRef for
        # comparison. This is to prevent duplicate namespaces with the same URI.
        if bound_namespace:
            bound_namespace = URIRef(bound_namespace)
        if bound_namespace and bound_namespace != namespace:

            # prefix already in use for different namespace
            if replace:
                self.store.bind(prefix, namespace)
            else:
                warnings.warn("Prefix ({}, {}) already defined, if want to replace it, set flag replace to True".format(
                    prefix if prefix else None, self.store.namespace(prefix)))
        else:
            bound_prefix = self.store.prefix(namespace)
            if bound_prefix is None:
                self.store.bind(prefix, namespace)
            elif bound_prefix == prefix:
                pass  # already bound
            else:
                if override or bound_prefix.startswith("_"):
                    self.store.bind(prefix, namespace)

    @staticmethod
    def check_uriref(text: Union[str, URI]) -> Optional[URIRef]:
        """
        Check if the input text is likely to be an URIRef and return None or URIRef
        """
        if isinstance(text, URIRef):
            return text
        if isinstance(text, URI):
            text = text.value
        if isinstance(text, str):
            text = text.strip()
            if URI_PATTERN.match(text.strip()):
                return URIRef(text)

    def split_uri(self, uri: str):
        """
        Overwrite rdflib's implementation which has a lot of issues
        """
        ns = ''
        for prefix, namespace in self.store.namespaces():
            if uri.startswith(namespace) and len(namespace) > len(ns):
                ns = namespace
        if ns:
            return ns, uri[len(ns):]
        raise SplitURIWithUnknownPrefix()

    def bind_for_master_config(self):
        """
        Bind must-have namespaces for master config, note RDF and XSD are already bound
        """
        self.bind('owl', OWL)
        self.bind('', 'http://isi.edu/default-ns/')

    def compute_qname(self, uri, generate=True):
        namespace, name = self.split_uri(uri)
        namespace = URIRef(namespace)
        prefix = self.store.prefix(namespace)
        if prefix is None:
            if not generate:
                raise Exception("No known prefix for %s and generate=False")
            num = 1
            while 1:
                prefix = "ns%s" % num
                if not self.store.namespace(prefix):
                    break
                num += 1
            self.bind(prefix, namespace)
        return prefix, namespace, name


if __name__ == '__main__':
    from rdflib import Graph
    nm = NamespaceManager(Graph())
    nm.bind('wdref', 'http://www.wikidata.org/reference/')
    print(nm.split_uri('http://www.wikidata.org/reference/355b56329b78db22be549dec34f2570ca61ca056'))
    print(nm.compute_qname('http://www.wikidata.org/reference/355b56329b78db22be549dec34f2570ca61ca056'))



"""
Example CLI module

Please DON'T import specific modules globally, import them in `run`.
Please DON'T initialize resource (e.g., variable) globally.
"""
from kgtk.cli_argparse import KGTKArgumentParser


def parser():
    """
    Initialize sub-parser.
    Parameters: https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser
    """
    return {
        'help': 'this is ticker',
        'description': 'it attaches current datetime to stdin'
    }


def add_arguments(parser: KGTKArgumentParser):
    """
    Parse arguments
    """
    parser.add_argument('-i', '--indent', action='count', default=0, help='indentation')


def run(indent):
    """
    Arguments here should be defined in `add_arguments` first.
    The return value (integer) will be the return code in shell. It will set to 0 if no value returns.
    You can either return a non-zero value to indicate error, or raise exceptions defined in kgtk.exceptions.
    """
    # import modules locally
    import sys
    from datetime import datetime
    sys.stdout.write(sys.stdin.read())
    sys.stdout.write('{}{}\n'.format('>' * indent, datetime.now()))



"""
Concatenate KGTK files.

TODO: Need KgtkWriterOptions
"""

from argparse import Namespace, SUPPRESS
from pathlib import Path
import sys
import typing

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Concatenate KGTK files.',
        'description': 'Concatenate two or more KGTK files, merging the columns appropriately. ' +
        'All files must be KGTK edge files or all files must be KGTK node files (unless overridden with --mode=NONE). ' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert cat --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderMode, KgtkReaderOptions
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.join.kgtkcat import KgtkCat
    from kgtk.utils.argparsehelpers import optional_bool
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_input_file(who="KGTK input files",
                          dest="input_files",
                          options=["-i", "--input-files"],
                          allow_list=True,
                          positional=True)
    parser.add_output_file()

    parser.add_argument("--output-format", dest="output_format", help="The file format (default=kgtk)", type=str,
                        choices=KgtkWriter.OUTPUT_FORMAT_CHOICES)

    parser.add_argument("--output-columns", dest="output_column_names",
                        metavar="NEW_COLUMN_NAME",
                        help=h("The list of new column names when renaming all columns."),
                        type=str, nargs='+')
    parser.add_argument("--old-columns", dest="old_column_names",
                        metavar="OLD_COLUMN_NAME",
                        help=h("The list of old column names for selective renaming."),
                        type=str, nargs='+')
    parser.add_argument("--new-columns", dest="new_column_names",
                        metavar="NEW_COLUMN_NAME",
                        help=h("The list of new column names for selective renaming."),
                        type=str, nargs='+')

    parser.add_argument("--no-output-header", dest="no_output_header", metavar="True|False",
                        help=h("When true, do not write a header to the output file (default=%(default)s)."),
                        type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument('--pure-python', dest='pure_python', metavar="True|False",
                        help="When True, use Python code. (default=%(default)s)",
                        type=optional_bool, nargs='?', const=True, default=KgtkCat.DEFAULT_PURE_PYTHON)

    parser.add_argument('--fast-copy-min-size', dest='fast_copy_min_size', type=int,
                        default=KgtkCat.DEFAULT_FAST_COPY_MIN_SIZE,
                        help='The minium number of bytes before using OS tools for fast copy (default=%(default)d).')

    parser.add_argument('--bash-command', dest='bash_command', type=str, default=KgtkCat.DEFAULT_BASH_COMMAND,
                        help=h("The bash command or its substitute. (default=%(default)s)"))

    parser.add_argument('--bzip2-command', dest='bzip2_command', type=str, default=KgtkCat.DEFAULT_BZIP2_COMMAND,
                        help=h("The bzip2 command or its substitute. (default=%(default)s)"))

    parser.add_argument('--cat-command', dest='cat_command', type=str, default=KgtkCat.DEFAULT_CAT_COMMAND,
                        help=h("The cat command or its substitute. (default=%(default)s)"))

    parser.add_argument('--gzip-command', dest='gzip_command', type=str, default=KgtkCat.DEFAULT_GZIP_COMMAND,
                        help=h("The gzip command or its substitute. (default=%(default)s)"))

    parser.add_argument('--tail-command', dest='tail_command', type=str, default=KgtkCat.DEFAULT_TAIL_COMMAND,
                        help=h("The tail command or its substitute. (default=%(default)s)"))

    parser.add_argument('--xz-command', dest='xz_command', type=str, default=KgtkCat.DEFAULT_XZ_COMMAND,
                        help=h("The xz command or its substitute. (default=%(default)s)"))

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser,
                                    mode_options=True,
                                    default_mode=KgtkReaderMode[parsed_shared_args._mode],
                                    expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)


def custom_progress() -> bool:
    return True  # We want to start a custom progress monitor.


def run(input_files: KGTKFiles,
        output_file: KGTKFiles,
        output_format: typing.Optional[str],

        output_column_names: typing.Optional[typing.List[str]],
        old_column_names: typing.Optional[typing.List[str]],
        new_column_names: typing.Optional[typing.List[str]],

        no_output_header: bool,
        pure_python: bool,
        fast_copy_min_size: int,

        bash_command: str,
        bzip2_command: str,
        cat_command: str,
        gzip_command: str,
        tail_command: str,
        xz_command: str,

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from kgtk.exceptions import KGTKException

    from kgtk.join.kgtkcat import KgtkCat
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions
    
    input_file_paths: typing.List[Path] = KGTKArgumentParser.get_input_file_list(input_files)
    output_file_path: Path = KGTKArgumentParser.get_output_file(output_file)
    
    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # print("cat running", file=error_file, flush=True) # ***

    # TODO: check that at most one input file is stdin?

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("--input-files %s" % " ".join((repr(str(input_file_path)) for input_file_path in input_file_paths)),
              file=error_file, flush=True)
        print("--output-file %s" % repr(str(output_file_path)), file=error_file, flush=True)
        if output_format is not None:
            print("--output-format %s" % repr(output_format), file=error_file, flush=True)
        if output_column_names is not None:
            print("--output-columns %s" % " ".join(output_column_names), file=error_file, flush=True)
        if old_column_names is not None:
            print("--old-columns %s" % " ".join(old_column_names), file=error_file, flush=True)
        if new_column_names is not None:
            print("--new-columns %s" % " ".join(new_column_names), file=error_file, flush=True)
        print("--no-output-header %s" % str(no_output_header), file=error_file, flush=True)
        print("--pure-python %s" % str(pure_python), file=error_file, flush=True)
        print("--fast-copy-min-size %d" % fast_copy_min_size, file=error_file, flush=True)
        print("--bash-commahd %s" % repr(bash_command), file=error_file, flush=True)
        print("--bzip2-commahd %s" % repr(bzip2_command), file=error_file, flush=True)
        print("--cat-commahd %s" % repr(cat_command), file=error_file, flush=True)
        print("--gzip-commahd %s" % repr(gzip_command), file=error_file, flush=True)
        print("--tail-commahd %s" % repr(tail_command), file=error_file, flush=True)
        print("--xz-commahd %s" % repr(xz_command), file=error_file, flush=True)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    # Check for consistent options.  argparse doesn't support this yet.
    if output_column_names is not None and len(output_column_names) > 0:
        if (old_column_names is not None and len(old_column_names) > 0) or \
           (new_column_names is not None and len(new_column_names) > 0):
            raise KGTKException("When --output-columns is used, --old-columns and --new-columns may not be used.")
    elif (old_column_names is not None and len(old_column_names) > 0) ^ \
         (new_column_names is not None and len(new_column_names) > 0):
        raise KGTKException("Both --old-columns and --new-columns must be used when either is used.")

    elif (old_column_names is not None and len(old_column_names) > 0) and \
         (new_column_names is not None and len(new_column_names) > 0):
        if len(old_column_names) != len(new_column_names):
            raise KGTKException("Both --old-columns and --new-columns must have the same number of columns.")
    try:
        kc: KgtkCat = KgtkCat(input_file_paths=input_file_paths,
                              output_path=output_file_path,
                              output_format=output_format,
                              output_column_names=output_column_names,
                              old_column_names=old_column_names,
                              new_column_names=new_column_names,
                              no_output_header=no_output_header,
                              pure_python=pure_python,
                              fast_copy_min_size=fast_copy_min_size,
                              bash_command=bash_command,
                              bzip2_command=bzip2_command,
                              cat_command=cat_command,
                              gzip_command=gzip_command,
                              tail_command=tail_command,
                              xz_command=xz_command,
                              reader_options=reader_options,
                              value_options=value_options,
                              error_file=error_file,
                              verbose=verbose,
                              very_verbose=very_verbose
        )
        
        kc.process()

        # print("cat done", file=error_file, flush=True) # ***
        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except KGTKException as e:
        raise
    except Exception as e:
        raise KGTKException(str(e))




"""Filter a KGTK file based on whether one or more records exist in a second
KGTK file with matching values for one or more fields.

Note: By default, this implementation builds im-memory sets of all the key
values in the second file (the filter file). Optionally, it will cache the
first file (the input file) instead.

Note: By default, input records are passed in order to the output file.  When
the input file is cached, the output records are order by key value (alpha
sort), then by input order.

TODO: Need KgtkWriterOptions
"""

from argparse import Namespace, SUPPRESS
import typing

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Filter a KGTK file by not matching records in a second KGTK file.',
        'description': 'Filter a KGTK file based on whether one or more records do not exist in a second KGTK file with matching values for one or more fields.' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert ifnotexists --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    from kgtk.iff.kgtkifexists import KgtkIfExists
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
    from kgtk.utils.argparsehelpers import optional_bool
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions
    
    _expert: bool = parsed_shared_args._expert

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_input_file(positional=True)
    parser.add_input_file(who="The KGTK file to filter against.",
                          options=["--filter-on", "--filter-file"], dest="filter_file", metavar="FILTER_FILE")
    parser.add_output_file()
    parser.add_output_file(who="The KGTK reject file for records that fail the filter.",
                           dest="reject_file",
                           options=["--reject-file"],
                           metavar="REJECT_FILE",
                           optional=True)

    parser.add_output_file(who="The KGTK file for filter records that matched at least one input record.",
                           dest="matched_filter_file",
                           options=["--matched-filter-file"],
                           metavar="MATCHED_FILTER_FILE",
                           optional=True)

    parser.add_output_file(who="The KGTK file for filter records that did not match any input records.",
                           dest="unmatched_filter_file",
                           options=["--unmatched-filter-file"],
                           metavar="UNMATCHED_FILTER_FILE",
                           optional=True)

    parser.add_output_file(who=h("The KGTK file for joined output records (EXPERIMENTAL)."),
                           dest="join_file",
                           options=["--join-file"],
                           metavar="JOIN_FILE",
                           optional=True)

    parser.add_argument(      "--input-keys", "--left-keys", dest="input_keys",
                              help="The key columns in the file being filtered (default=None).", nargs='*')

    parser.add_argument(      "--filter-keys", "--right-keys", dest="filter_keys",
                              help="The key columns in the filter-on file (default=None).", nargs='*')

    parser.add_argument(      "--cache-input", dest="cache_input", metavar="True|False",
                              help="Cache the input file instead of the filter keys (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--preserve-order", dest="preserve_order", metavar="True|False",
                              help="Preserve record order when cacheing the input file. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--presorted", dest="presorted", metavar="True|False",
                              help="When True, assume that the input and filter files are both presorted.  Use a merge-style algorithm that does not require caching either file. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--field-separator", dest="field_separator",
                              help=h("Separator for multifield keys (default=%(default)s)"),
                              default=KgtkIfExists.FIELD_SEPARATOR_DEFAULT)

    parser.add_argument(      "--left-join", dest="left_join",  metavar="True|False",
                              help=h("When True, Include all input records in the join (EXPERIMENTAL). (default=%(default)s)."),
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--right-join", dest="right_join",  metavar="True|False",
                              help=h("When True, Include all filter records in the join (EXPERIMENTAL). (default=%(default)s)."),
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--input-prefix", dest="input_prefix",
                              help=h("Input file column name prefix for joins (EXPERIMENTAL). (default=%(default)s)"))

    parser.add_argument(      "--filter-prefix", dest="filter_prefix",
                              help=h("Filter file column name prefix for joins (EXPERIMENTAL). (default=%(default)s)"))

    parser.add_argument(      "--join-output", dest="join_output",  metavar="True|False",
                              help=h("When True, send the join records to the main output (EXPERIMENTAL). (default=%(default)s)."),
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--right-join-first", dest="right_first",  metavar="True|False",
                              help=h("When True, send the filter record to join output before the first matching input record. " +
                              " Otherwise, send the first matching input record, then the filter record, then othe rmatching input records. " +
                              "(EXPERIMENTAL). (default=%(default)s)."),
                              type=optional_bool, nargs='?', const=True, default=False)

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser,
                                    mode_options=True,
                                    default_mode=KgtkReaderMode[parsed_shared_args._mode],
                                    expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, who="input", expert=_expert, defaults=False)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, who="filter", expert=_expert, defaults=False)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(input_file: KGTKFiles,
        filter_file: KGTKFiles,
        output_file: KGTKFiles,
        reject_file: KGTKFiles,
        matched_filter_file: KGTKFiles,
        unmatched_filter_file: KGTKFiles,
        join_file: KGTKFiles,

        input_keys: typing.Optional[typing.List[str]],
        filter_keys: typing.Optional[typing.List[str]],
        
        cache_input: bool = False,
        preserve_order: bool = False,
        presorted: bool = False,

        field_separator: typing.Optional[str] = None,

        left_join: bool = False,
        right_join: bool = False,
        input_prefix: typing.Optional[str] = None,
        filter_prefix: typing.Optional[str] = None,
        join_output: bool = False,
        right_first: bool = False,

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from pathlib import Path
    import sys

    from kgtk.exceptions import KGTKException
    from kgtk.iff.kgtkifexists import KgtkIfExists
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    input_kgtk_file: Path = KGTKArgumentParser.get_input_file(input_file)
    filter_kgtk_file: Path = KGTKArgumentParser.get_input_file(filter_file, who="KGTK filter file")
    output_kgtk_file: Path = KGTKArgumentParser.get_output_file(output_file)
    reject_kgtk_file: typing.Optional[Path] = KGTKArgumentParser.get_optional_output_file(reject_file, who="KGTK reject file")
    matched_filter_kgtk_file: typing.Optional[Path] = KGTKArgumentParser.get_optional_output_file(matched_filter_file, who="KGTK matched filter file")
    unmatched_filter_kgtk_file: typing.Optional[Path] = KGTKArgumentParser.get_optional_output_file(unmatched_filter_file, who="KGTK unmatched filter file")
    join_kgtk_file: typing.Optional[Path] = KGTKArgumentParser.get_optional_output_file(join_file, who="KGTK join file")

    if (str(input_kgtk_file) == "-" and str(filter_kgtk_file) == "-"):
        raise KGTKException("My not use stdin for both --input-file and --filter-on files.")

    field_separator = KgtkIfExists.FIELD_SEPARATOR_DEFAULT if field_separator is None else field_separator

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    input_reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs, who="input", fallback=True)
    filter_reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs, who="filter", fallback=True)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("input: %s" % (str(input_kgtk_file) if input_kgtk_file is not None else "-"), file=error_file)
        print("--filter-file=%s" % (str(filter_kgtk_file) if filter_kgtk_file is not None else "-"), file=error_file)
        print("--output-file=%s" % (str(output_kgtk_file) if output_kgtk_file is not None else "-"), file=error_file)
        if reject_kgtk_file is not None:
            print("--reject-file=%s" % str(reject_kgtk_file), file=error_file)
        if matched_filter_kgtk_file is not None:
            print("--matched-filter-file=%s" % str(matched_filter_kgtk_file), file=error_file)
        if unmatched_filter_kgtk_file is not None:
            print("--unmatched-filter-file=%s" % str(unmatched_filter_kgtk_file), file=error_file)
        if join_kgtk_file is not None:
            print("--join-file=%s" % str(join_kgtk_file), file=error_file)
        if input_keys is not None:
            print("--input-keys=%s" % " ".join(input_keys), file=error_file)
        if filter_keys is not None:
            print("--filter-keys=%s" % " ".join(filter_keys), file=error_file)
        print("--cache-input=%s" % str(cache_input), file=error_file)
        print("--preserve-order=%s" % str(preserve_order), file=error_file)
        print("--presorted=%s" % str(presorted), file=error_file)
        print("--field-separator='%s'" % repr(field_separator), file=error_file)
        print("--left-join=%s" % str(left_join), file=error_file)
        print("--right-join=%s" % str(right_join), file=error_file)
        if input_prefix is not None:
            print("--input-prefix=%s" % repr(input_prefix), file=error_file)
        if filter_prefix is not None:
            print("--filter-prefix=%s" % repr(filter_prefix), file=error_file)
        print("--join-output=%s" % str(join_output), file=error_file)
        print("--right-join-first=%s" % str(right_first), file=error_file)
        input_reader_options.show(out=error_file, who="input")
        filter_reader_options.show(out=error_file, who="filter")
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    try:
        ie: KgtkIfExists = KgtkIfExists(
            input_file_path=input_kgtk_file,
            input_keys=input_keys,
            filter_file_path=filter_kgtk_file,
            filter_keys=filter_keys,
            output_file_path=output_kgtk_file,
            reject_file_path=reject_kgtk_file,
            matched_filter_file_path=matched_filter_kgtk_file,
            unmatched_filter_file_path=unmatched_filter_kgtk_file,
            join_file_path=join_kgtk_file,
            left_join=left_join,
            right_join=right_join,
            input_prefix=input_prefix,
            filter_prefix=filter_prefix,
            join_output=join_output,
            right_first=right_first,
            invert=True,
            cache_input=cache_input,
            preserve_order=preserve_order,
            presorted=presorted,
            field_separator=field_separator,
            input_reader_options=input_reader_options,
            filter_reader_options=filter_reader_options,
            value_options=value_options,
            error_file=error_file,
            verbose=verbose,
            very_verbose=very_verbose,
        )
        
        ie.process()

        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))




files_config = {
    "alias_de": "aliases.de.tsv.gz",
    "alias": "aliases.en.tsv.gz",
    "alias_es": "aliases.es.tsv.gz",
    "alias_fr": "aliases.fr.tsv.gz",
    "alias_it": "aliases.it.tsv.gz",
    "alias_nl": "aliases.nl.tsv.gz",
    "alias_pl": "aliases.pl.tsv.gz",
    "alias_pt": "aliases.pt.tsv.gz",
    "alias_ru": "aliases.ru.tsv.gz",
    "alias_sv": "aliases.sv.tsv.gz",
    "alias_all": "aliases.tsv.gz",
    "alias_zhcn": "aliases.zh-cn.tsv.gz",
    "all": "all.tsv.gz",
    "commondsmedia": "claims.commonsMedia.tsv.gz",
    "external_id": "claims.external-id.tsv.gz",
    "geo_shape": "claims.geo-shape.tsv.gz",
    "globe_coordinate": "claims.globe-coordinate.tsv.gz",
    "math": "claims.math.tsv.gz",
    "monolingualtext": "claims.monolingualtext.tsv.gz",
    "musical_notation": "claims.musical-notation.tsv.gz",
    "claims_other": "claims.other.tsv.gz",
    "quantity": "claims.quantity.tsv.gz",
    "string": "claims.string.tsv.gz",
    "tabular_data": "claims.tabular-data.tsv.gz",
    "time": "claims.time.tsv.gz",
    "claims": "claims.tsv.gz",
    "url": "claims.url.tsv.gz",
    "wikibase_form": "claims.wikibase-form.tsv.gz",
    "item": "claims.wikibase-item.tsv.gz",
    "wikibase_lexeme": "claims.wikibase-lexeme.tsv.gz",
    "wikibase_property": "claims.wikibase-property.tsv.gz",
    "wikibase_sense": "claims.wikibase-sense.tsv.gz",
    "dwd_isa": "derived.dwd_isa.tsv.gz",
    "isastar": "derived.isastar.tsv.gz",
    "isa": "derived.isa.tsv.gz",
    "p279star": "derived.P279star.tsv.gz",
    "p279": "derived.P279.tsv.gz",
    "p31p279star": "derived.P31P279star.tsv.gz",
    "p31": "derived.P31.tsv.gz",
    "description_de": "descriptions.de.tsv.gz",
    "description": "descriptions.en.tsv.gz",
    "description_es": "descriptions.es.tsv.gz",
    "description_fr": "descriptions.fr.tsv.gz",
    "description_it": "descriptions.it.tsv.gz",
    "description_nl": "descriptions.nl.tsv.gz",
    "description_pl": "descriptions.pl.tsv.gz",
    "description_pt": "descriptions.pt.tsv.gz",
    "description_ru": "descriptions.ru.tsv.gz",
    "description_sv": "descriptions.sv.tsv.gz",
    "description_all": "descriptions.tsv.gz",
    "description_zhcn": "descriptions.zh-cn.tsv.gz",
    "label_de": "labels.de.tsv.gz",
    "label": "labels.en.tsv.gz",
    "KGTK_LABEL_FILE": "labels.en.tsv.gz",
    "label_es": "labels.es.tsv.gz",
    "label_fr": "labels.fr.tsv.gz",
    "label_it": "labels.it.tsv.gz",
    "label_nl": "labels.nl.tsv.gz",
    "label_pl": "labels.pl.tsv.gz",
    "label_pt": "labels.pt.tsv.gz",
    "label_ru": "labels.ru.tsv.gz",
    "label_sv": "labels.sv.tsv.gz",
    "label_all": "labels.tsv.gz",
    "label_zhcn": "labels.zh-cn.tsv.gz",
    "in_degree": "metadata.in_degree.tsv.gz",
    "out_degree": "metadata.out_degree.tsv.gz",
    "pagerank_directed": "metadata.pagerank.directed.tsv.gz",
    "pagerank_undirected": "metadata.pagerank.undirected.tsv.gz",
    "datatypes_augmented": "metadata.property.datatypes.augmented.tsv.gz",
    "datatypes": "metadata.property.datatypes.tsv.gz",
    "types": "metadata.types.tsv.gz",
    "qualifiers_commonsmedia": "qualifiers.commonsMedia.tsv.gz",
    "qualifiers_external_id": "qualifiers.external-id.tsv.gz",
    "qualifiers_geo_shape": "qualifiers.geo-shape.tsv.gz",
    "qualifiers_globe_coordinate": "qualifiers.globe-coordinate.tsv.gz",
    "qualifiers_math": "qualifiers.math.tsv.gz",
    "qualifiers_monolingual_text": "qualifiers.monolingualtext.tsv.gz",
    "qualifiers_musical_notation": "qualifiers.musical-notation.tsv.gz",
    "qualifiers_quantity": "qualifiers.quantity.tsv.gz",
    "qualifiers_string": "qualifiers.string.tsv.gz",
    "qualifiers_tabular_data": "qualifiers.tabular-data.tsv.gz",
    "qualifiers_time": "qualifiers.time.tsv.gz",
    "qualifiers": "qualifiers.tsv.gz",
    "qualifiers_url": "qualifiers.url.tsv.gz",
    "qualifiers_wikibase_form": "qualifiers.wikibase-form.tsv.gz",
    "qualifiers_wikibase_item": "qualifiers.wikibase-item.tsv.gz",
    "qualifiers_wikibase_lexeme": "qualifiers.wikibase-lexeme.tsv.gz",
    "qualifiers_wikibase_property": "qualifiers.wikibase-property.tsv.gz",
    "qualifiers_wikibase_sense": "qualifiers.wikibase-sense.tsv.gz",
    "sitelinks_en": "sitelinks.en.tsv.gz",
    "sitelinks_qualifiers_en": "sitelinks.qualifiers.en.tsv.gz",
    "sitelinks_qualifiers": "sitelinks.qualifiers.tsv.gz",
    "sitelinks": "sitelinks.tsv.gz"
}



import shutil
import unittest
import tempfile
import pandas as pd
from kgtk.cli_entry import cli_entry
from kgtk.exceptions import KGTKException

class TestKGTKImportConceptNet(unittest.TestCase):
    def setUp(self) -> None:
        self.temp_dir=tempfile.mkdtemp()

    def tearDown(self) -> None:
        shutil.rmtree(self.temp_dir)

    def test_kgtk_import_conceptnet(self):
        cli_entry("kgtk", "import-conceptnet", "--english_only", "-i", "data/conceptnet.csv", "-o", f'{self.temp_dir}/conceptnet.tsv')

        df = pd.read_csv(f'{self.temp_dir}/conceptnet.tsv', sep='\t')

        self.assertEqual(len(df), 77) # Make sure that there are 77 rows with English nodes out of the total 100

        self.assertEqual(df['node2'][0], '/c/en/1') # Make sure that the first node2 is the expected one

        self.assertEqual(len(df.columns), 9)

        relations = df['relation'].unique()




import shutil
import unittest
import tempfile
import pandas as pd
from kgtk.cli_entry import cli_entry
from kgtk.exceptions import KGTKException

class TestImportFrameNet(unittest.TestCase):
    def setUp(self) -> None:
        self.temp_dir=tempfile.mkdtemp()

    def tearDown(self) -> None:
        shutil.rmtree(self.temp_dir)

    def test_kgtk_import_framenet(self):
        cli_entry("kgtk", "import_framenet", "-o", f'{self.temp_dir}/framenet.tsv')

        df = pd.read_csv(f'{self.temp_dir}/framenet.tsv', sep='\t', na_filter=False)

        self.assertEqual(len(df.columns), 9) # Make sure that the amount of columns is as expected

        self.assertEqual(len(df), 29873) # Make sure that the amount of rows is as expected




"""Count records or non-empty values per column.

This is a simple command that illustrates several aspects of building
a KGTK command.  The following features are illustrated:

* Using the KGTK command line parser's extended features.
  * Adding command alias.
  * Adding an option with a different default for the command alias.
  * Adding an expert option.
* Reading a KGTK input file.
* Writing a KGTK output file.
* Writing non-KGTK output to stdout.
* Writing progress feedback to etderr.
"""

# We minimize the imports here because every additional global import
# slows down `kgtk` command initialization.
from argparse import Namespace, SUPPRESS
import typing

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

# Decine the name of the command and its alias.
COUNT_COMMAND: str = "count"
WC_COMMAND: str = "wc"

# Default option values:
DEFAULT_COUNT_RECORDS: bool = False
DEFAULT_COUNT_RECORDS_WC: bool = True
DEFAULT_COUNT_PROPERTY: str = "count"

def parser():
    return {
        'aliases': [ WC_COMMAND ],
        'help': 'Count records or non-empty values per column.',
        'description': 'Count the number of records in a KGTK file, excluding the header record, ' +
        'or count the number of non-empty values per column.  Note:  not non-empty unique values, ' +
        'that is what `kgtk unique` does.' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert lift --help'
    }

def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """                                                                                                                                                               
    Parse arguments                                                                                                                                                   
    Args:                                                                                                                                                             
        parser (argparse.ArgumentParser)                                                                                                                              
    """
    # Import modules thay we will use when declaring arguments.
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
    from kgtk.utils.argparsehelpers import optional_bool
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    # These special shared aruments inticate whether the `--expert` option
    # was supplied and the command name that was used.
    _expert: bool = parsed_shared_args._expert
    _command: str = parsed_shared_args._command

    # This helper function makes it easy to suppress options from the help
    # message unless `--expert` has bee asserted.  The options are still
    # there, and initialize what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    # Add the primary input and output files without special features.
    parser.add_input_file()
    parser.add_output_file()

    # The default value for this option depends upon the command used.
    parser.add_argument('-l', '--lines', dest="count_records", metavar="True/False",
                        help="If true, count records and print a single number to stdout. " +
                        "If false, count non-empty values per column and produce a simple KGTK output file. (default=%(default)s).",
                        type=optional_bool, nargs='?', const=True,
                        default=DEFAULT_COUNT_RECORDS_WC if _command == WC_COMMAND else DEFAULT_COUNT_RECORDS)

    # This is an expert option.  It will not show up on `--help` without `--expert`:
    parser.add_argument(      "--count-property", dest="count_property",
                              help=h("The property used for column count output edges. (default=%(default)s)."),
                              default=DEFAULT_COUNT_PROPERTY)

    # Add the standard debugging arguments and the KgtkReader and KgtkValue
    # options.
    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser,
                                    mode_options=True,
                                    default_mode=KgtkReaderMode[parsed_shared_args._mode],
                                    expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)


# This is our irimary entry point after command parsing.  Note the special
# KGTKFiles type used for input and output files that were declared using the
# extended argument parser's `add_input_file()` and `add_output_file()`
# methods.
def run(input_file: KGTKFiles,
        output_file: KGTKFiles,

        count_records: bool = DEFAULT_COUNT_RECORDS,
        count_property: str = DEFAULT_COUNT_PROPERTY,

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:

    # Import the modules that we need for processing:
    import sys

    from kgtk.exceptions import KGTKException
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.io.kgtkwriter import KgtkWriter

    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    # Perform the final parsing steps on the primary input and output files.
    # The naming convention is unfortunate;  these should have been called
    # `input_kgtk_path` and `output_kgtk_path`.
    input_kgtk_file: Path = KGTKArgumentParser.get_input_file(input_file)
    output_kgtk_file: Path = KGTKArgumentParser.get_output_file(output_file)

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the KGTKReader and KgtkValue option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Show the final option values for debugging and documentation.
    if show_options:
        print("--input-file=%s" % str(input_kgtk_file), file=error_file, flush=True)
        print("--output-file=%s" % str(output_kgtk_file), file=error_file, flush=True)
        print("--lines=%s" % repr(count_records),  file=error_file, flush=True)
        print("--count-property=%s" % repr(count_property),  file=error_file, flush=True)

        input_reader_options.show(out=error_file, who="input")
        label_reader_options.show(out=error_file, who="label")
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    # Prepare to catch SystemExit and Exception.
    try:
        
        # Open the primary input file:
        if verbose:
            print("Opening the input file %s" % str(input_kgtk_file), file=error_file, flush=True)
        kr: KgtkReader = KgtkReader.open(input_kgtk_file,
                                         options=reader_options,
                                         value_options = value_options,
                                         error_file=error_file,
                                         verbose=verbose,
                                         very_verbose=very_verbose,
        )

        # We have two execution paths, depending on count_records:
        row: typing.List[str]
        if count_records:
            # Count the number of data records in the input file.
            record_count: int = 0
            for row in kr:
                record_count += 1

            if str(output_kgtk_file) == "-":
                # Send the record count to standard output:
                print("%d" % record_count, file=sys.stdout, flush=True)

            else:
                with open(output_kgtk_file, "w") as outf:
                    # Send the record count to a file:
                    print("%d" % record_count, file=outf)

        else:
            # Open a KgtkWriter file for the primary output file:
            if verbose:
                print("Opening the output file %s" % str(output_kgtk_file), file=error_file, flush=True)
            kw: KgtkWriter = KgtkWriter.open(["node1", "label", "node2" ],
                                             output_kgtk_file,
                                             verbose=verbose,
                                             very_verbose=very_verbose,
                                 )
            # Count the number of non-empty cells for each column:
            record_counts: typing.List[int] = [ 0 for idx in range(kr.column_count) ]
            idx: int
            for row in kr:
                item: str
                for idx, item in enumerate(row):
                    if len(item) > 0:
                        record_counts[idx] += 1

            # Send the results to the primary output file as a series of edges.
            count: int
            for idx, count in enumerate(record_counts):
                kw.write([ kr.column_names[idx], count_property, str(record_counts[idx]) ])

            # Close the primary output file.
            kw.close()            

        # Close the primary input file.
        kr.close()
        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))



from argparse import Namespace, SUPPRESS
import typing

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles


def parser():
    return {
        'help': 'Find connected components in a Graph.',
        'description': 'Find all the connected components in an undirected or directed Graph.' +
                       '\n\nAdditional options are shown in expert help.\nkgtk --expert connected-components --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    from kgtk.gt.connected_components import ConnectedComponents
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
    from kgtk.utils.enumnameaction import EnumLowerNameAction

    _expert: bool = parsed_shared_args._expert

    parser.add_input_file(positional=True, who="The KGTK file to find connected components in.")
    parser.add_output_file()

    parser.add_argument("--properties", action="store", type=str, dest="properties",
                        help=' A comma separated list of properties to traverse while finding connected components, '
                             'by default all properties will be considered',
                        default='')
    parser.add_argument('--undirected', action='store_true', dest="undirected",
                        help="Specify if the input graph is undirected, default FALSE")
    parser.add_argument('--strong', action='store_true', dest="strong",
                        help="Treat graph as directed or not, independent of its actual directionality.")


    parser.add_argument("--cluster-name-method", dest="cluster_name_method",
                        help="Determine the naming method for clusters. (default=%(default)s)",
                        type=ConnectedComponents.Method, action=EnumLowerNameAction,
                        default=ConnectedComponents.DEFAULT_CLUSTER_NAME_METHOD)
    
    parser.add_argument("--cluster-name-separator", dest="cluster_name_separator",
                        help="Specify the separator to be used in cat and hash cluster name methods. (default=%(default)s)",
                        default=ConnectedComponents.DEFAULT_CLUSTER_NAME_SEPARATOR)

    parser.add_argument("--cluster-name-prefix", dest="cluster_name_prefix",
                        help="Specify the prefix to be used in the prefixed and hash cluster name methods. (default=%(default)s)",
                        default=ConnectedComponents.DEFAULT_CLUSTER_NAME_PREFIX)

    parser.add_argument("--cluster-name-zfill", dest="cluster_name_zfill", type=int,
                        help="Specify the zfill to be used in the numbered and prefixed cluster name methods. (default=%(default)s)",
                        default=ConnectedComponents.DEFAULT_CLUSTER_NAME_ZFILL)

    parser.add_argument("--minimum-cluster-size", dest="minimum_cluster_size", type=int,
                        help="Specify the minimum cluster size. (default=%(default)s)",
                        default=ConnectedComponents.DEFAULT_MINIMUM_CLUSTER_SIZE)


    # CMR: The folowing options aren't used.  Is the intent to support them?
    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser,
                                    mode_options=True,
                                    default_mode=KgtkReaderMode[parsed_shared_args._mode],
                                    expert=_expert)
    KgtkReaderOptions.add_arguments(parser,
                                    mode_options=True,
                                    default_mode=KgtkReaderMode[parsed_shared_args._mode],
                                    who="input",
                                    expert=_expert,
                                    defaults=False)
    # KgtkValueOptions.add_arguments(parser, expert=_expert)


def run(input_file: KGTKFiles,
        output_file: KGTKFiles,

        properties: str = '',
        undirected: bool = False,
        strong: bool = False,

        # The following have been modified to postpone importing gtaph_tools.
        # ClusterComponents cann't be referenced here.
        cluster_name_method: typing.Optional[typing.Any] = None,
        cluster_name_separator: typing.Optional[str] = None,
        cluster_name_prefix: typing.Optional[str] = None,
        cluster_name_zfill: typing.Optional[int] = None,
        minimum_cluster_size: typing.Optional[int] = None,

        **kwargs  # Whatever KgtkFileOptions and KgtkValueOptions want.
        ) -> int:
    from pathlib import Path
    
    from kgtk.exceptions import KGTKException
    from kgtk.gt.connected_components import ConnectedComponents
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions

    input_kgtk_file: Path = KGTKArgumentParser.get_input_file(input_file)
    output_kgtk_file: Path = KGTKArgumentParser.get_output_file(output_file)

    # It's OK to mention ClusterComponents here.
    cluster_name_method_x: ConnectedComponents.Method = \
        cluster_name_method if cluster_name_method is not None else ConnectedComponents.DEFAULT_CLUSTER_NAME_METHOD
    cluster_name_separator = ConnectedComponents.DEFAULT_CLUSTER_NAME_SEPARATOR if cluster_name_separator is None else cluster_name_separator
    cluster_name_prefix = ConnectedComponents.DEFAULT_CLUSTER_NAME_PREFIX if cluster_name_prefix is None else cluster_name_prefix
    cluster_name_zfill = ConnectedComponents.DEFAULT_CLUSTER_NAME_ZFILL if cluster_name_zfill is None else cluster_name_zfill
    minimum_cluster_size = ConnectedComponents.DEFAULT_MINIMUM_CLUSTER_SIZE if minimum_cluster_size is None else minimum_cluster_size

    cc: ConnectedComponents = ConnectedComponents(input_file_path=input_kgtk_file,
                                                  output_file_path=output_kgtk_file,
                                                  properties=properties,
                                                  undirected=undirected,
                                                  strong=strong,
                                                  cluster_name_method=cluster_name_method_x,
                                                  cluster_name_separator=cluster_name_separator,
                                                  cluster_name_prefix=cluster_name_prefix,
                                                  cluster_name_zfill=cluster_name_zfill,
                                                  minimum_cluster_size=minimum_cluster_size,
    )

    try:
        cc.process()
        return 0
    except Exception as e:
        raise KGTKException(str(e))



"""
Rename KGTK file columns (while copying)

TODO: Need KgtkWriterOptions
"""

from argparse import Namespace, SUPPRESS
import typing

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Rename KGTK file columns.',
        'description': 'This command renames one or more columns in a KGTK file. ' +
        '\n\nRename all columns using --output-columns newname1 newname2 ... ' +
        '\nRename selected columns using --old-columns and --new-columns ' +
        '\n\nThe column names are listed seperately for each option, do not quote them as a group, e.g. ' +
        '\nkgtk rename_columns --old-columns oldname1 oldname2 --new-columns newname1 nsewname2' +
        '\n\nThe input filename must come before --output-columns, --old-columns, or --new-columns. ' +
        '\nIf no input filename is provided, the default is to read standard input. ' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert rename-columns --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_input_file(positional=True)
    parser.add_output_file()

    parser.add_argument(      "--output-format", dest="output_format", help=h("The file format (default=kgtk)"), type=str)

    parser.add_argument(      "--output-columns", dest="output_column_names",
                              metavar="NEW_COLUMN_NAME",
                              help="The list of new column names when renaming all columns.",
                              type=str, nargs='+')
    parser.add_argument(      "--old-columns", dest="old_column_names",
                              metavar="OLD_COLUMN_NAME",
                              help="The list of old column names for selective renaming.",
                              type=str, nargs='+', action="append", default=list())
    parser.add_argument(      "--new-columns", dest="new_column_names",
                              metavar="NEW_COLUMN_NAME",
                              help="The list of new column names for selective renaming.",
                              type=str, nargs='+', action="append", default=list())

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser,
                                    mode_options=True,
                                    default_mode=KgtkReaderMode.NONE,
                                    expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(input_file: KGTKFiles,
        output_file: KGTKFiles,
        output_format: typing.Optional[str],

        output_column_names: typing.Optional[typing.List[str]],
        old_column_names: typing.Optional[typing.List[typing.List[str]]],
        new_column_names: typing.Optional[typing.List[typing.List[str]]],

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from pathlib import Path
    import sys

    from kgtk.exceptions import KGTKException
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.join.kgtkcat import KgtkCat
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    input_file_path: Path = KGTKArgumentParser.get_input_file(input_file)
    output_file_path: Path = KGTKArgumentParser.get_output_file(output_file)

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Condense the old and new columns names lists.
    old_column_names_compact: typing.List[str] = list()
    column_name_list: typing.List[str]
    column_name: str
    if old_column_names is not None:
        for column_name_list in old_column_names:
            for column_name in column_name_list:
                old_column_names_compact.append(column_name)

    new_column_names_compact: typing.List[str] = list()
    if new_column_names is not None:
        for column_name_list in new_column_names:
            for column_name in column_name_list:
                new_column_names_compact.append(column_name)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("--input-file=%s" % str(input_file_path), file=error_file, flush=True)
        print("--output-file=%s" % str(output_file_path), file=error_file, flush=True)
        if output_format is not None:
            print("--output-format=%s" % output_format, file=error_file, flush=True)
        if output_column_names is not None:
            print("--output-columns %s" % " ".join(output_column_names), file=error_file, flush=True)
        if len(old_column_names_compact) > 0:
            print("--old-columns %s" % " ".join(old_column_names_compact), file=error_file, flush=True)
        if len(new_column_names_compact) > 0:
            print("--new-columns %s" % " ".join(new_column_names_compact), file=error_file, flush=True)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    # Check for consistent options.  argparse doesn't support this yet.
    if output_column_names is not None and len(output_column_names) > 0:
        if (old_column_names is not None and len(old_column_names) > 0) or \
           (new_column_names is not None and len(new_column_names) > 0):
            raise KGTKException("When --output-columns is used, --old-columns and --new-columns may not be used.")
    elif old_column_names is not None and len(old_column_names) > 0:
        if new_column_names is None or len(new_column_names) == 0:
            raise KGTKException("Both --old-columns and --new-columns must be used when either is used.")
        if len(old_column_names) != len(new_column_names):
            raise KGTKException("Both --old-columns and --new-columns must have the same number of columns.")
    elif new_column_names is not None and len(new_column_names) > 0:
        if old_column_names is None or len(old_column_names) == 0:
            raise KGTKException("Both --old-columns and --new-columns must be used when either is used.")
    else:
        raise KGTKException("You must specify --output-columns or both of --old-columns and --new-columns.")

    try:
        kc: KgtkCat = KgtkCat(input_file_paths=[input_file_path],
                              output_path=output_file_path,
                              output_format=output_format,
                              output_column_names=output_column_names,
                              old_column_names=old_column_names_compact,
                              new_column_names=new_column_names_compact,
                              reader_options=reader_options,
                              value_options=value_options,
                              error_file=error_file,
                              verbose=verbose,
                              very_verbose=very_verbose
        )
        
        kc.process()

        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))




"""
Reorder KGTK file columns (while copying)

TODO: Need KgtkWriterOptions

TODO: Make --as-columns modify the preceeding --columns such that
--as-columns can be applied selectively.
"""

from argparse import Namespace, SUPPRESS
import typing

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

REORDER_COLUMNS_COMMAND: str = 'reorder-columns'
SELECT_COLUMNS_COMMAND: str = 'select-columns'

def parser():
    return {
        'aliases': [ SELECT_COLUMNS_COMMAND ],
        'help': 'Reorder KGTK file columns.',
        'description': 'This command reorders one or more columns in a KGTK file. ' +
        '\n\nReorder all columns using --columns col1 col2' +
        '\nReorder selected columns using --columns col1 col2 ... coln-1 coln' +
        '\nMove a column to the front: --columns col ...' +
        '\nMove a column to the end: --columns ... col' +
        '\nExtract named columns, omitting the rest: --columns col1 col2 --trim' +
        '\nMove a range of columns: --columns coln .. colm ...' +
        '\nIf no input filename is provided, the default is to read standard input. ' +
        '\n\nAdditional options are shown in expert help.\nkgtk --expert reorder-columns --help'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    # import modules locally
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
    from kgtk.utils.argparsehelpers import optional_bool
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert
    _command: str = parsed_shared_args._command

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.
    def h(msg: str)->str:
        if _expert:
            return msg
        else:
            return SUPPRESS

    parser.add_input_file()
    parser.add_output_file()

    parser.add_argument(      "--output-format", dest="output_format", help=h("The file format (default=kgtk)"), type=str)

    parser.add_argument('-c', "--columns", "--column", dest="column_names_list", required=True, nargs='+', action="append", default=list(),
                              metavar="COLUMN_NAME",
                              help="The list of reordered column names, optionally containing '...' for column names not explicitly mentioned.")

    parser.add_argument(      "--as", "--as-columns", "--as-column", dest="as_column_names_list", nargs='+', action="append", default=list(),
                              metavar="COLUMN_NAME",
                              help="Replacement column names.")

    parser.add_argument(      "--trim", dest="omit_remaining_columns",
                              help="If true, omit unmentioned columns. (default=%(default)s).",
                              metavar="True|False",
                              type=optional_bool, nargs='?', const=True, default=(_command == SELECT_COLUMNS_COMMAND))

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser,
                                    mode_options=True,
                                    default_mode=KgtkReaderMode.NONE,
                                    expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(input_file: KGTKFiles,
        output_file: KGTKFiles,
        output_format: typing.Optional[str],

        column_names_list: typing.List[typing.List[str]],
        as_column_names_list: typing.List[typing.List[str]],

        omit_remaining_columns: bool,

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from pathlib import Path
    import sys
    
    from kgtk.exceptions import KGTKException
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions


    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    input_kgtk_file: Path = KGTKArgumentParser.get_input_file(input_file)
    output_kgtk_file: Path = KGTKArgumentParser.get_output_file(output_file)

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    # Condense the old and new columns names lists.
    column_names: typing.List[str] = list()
    column_name_list: typing.List[str]
    column_name: str
    if column_names_list is not None:
        for column_name_list in column_names_list:
            for column_name in column_name_list:
                column_names.append(column_name)

    as_column_names: typing.List[str] = list()
    if as_column_names_list is not None:
        for column_name_list in as_column_names_list:
            for column_name in column_name_list:
                as_column_names.append(column_name)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("--input-file=%s" % str(input_kgtk_file), file=error_file, flush=True)
        print("--output-file=%s" % str(output_kgtk_file), file=error_file, flush=True)
        if output_format is not None:
            print("--output-format=%s" % output_format, file=error_file, flush=True)
        print("--columns %s" % " ".join(column_names), file=error_file, flush=True)
        if len(as_column_names) > 0:
            print("--as-columns %s" % " ".join(as_column_names), file=error_file, flush=True)
        print("--trim=%s" % str(omit_remaining_columns), file=error_file, flush=True)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)

    # Check for consistent options.  argparse doesn't support this yet.
    if len(as_column_names) > 0 and len(as_column_names) != len(column_names):
        raise KGTKException("Both --columns and --as-columns must have the same number of columns when --as-columns is used.")

    try:

        if verbose:
            print("Opening the input file %s" % str(input_kgtk_file), file=error_file, flush=True)
        kr = KgtkReader.open(input_kgtk_file,
                             options=reader_options,
                             value_options = value_options,
                             error_file=error_file,
                             verbose=verbose,
                             very_verbose=very_verbose,
        )

        remaining_names: typing.List[str] = kr.column_names.copy()
        reordered_names: typing.List[str] = [ ]
        save_reordered_names: typing.Optional[typing.List[str]] = None

        ellipses: str = "..." # All unmentioned columns
        ranger: str = ".." # All columns between two columns.

        saw_ranger: bool = False
        idx: int
        column_name: str
        for idx, column_name in enumerate(column_names):
            if column_name == ellipses:
                if len(as_column_names) > 0:
                    raise  KGTKException("The elipses operator ('...') may not appear when --as-columns is specified.")
                if save_reordered_names is not None:
                    raise KGTKException("Elipses may appear only once")

                if saw_ranger:
                    raise KGTKException("Elipses may not appear directly after a range operator ('..').")

                save_reordered_names = reordered_names
                reordered_names = [ ]
                continue

            if column_name == ranger:
                if len(reordered_names) == 0:
                    raise KGTKException("The column range operator ('..') may not appear without a preceeding column name.")
                if len(as_column_names) > 0:
                    raise  KGTKException("The column range operator ('..') may not appear when --as-columns is specified.")
                saw_ranger = True
                continue

            if column_name not in kr.column_names:
                raise KGTKException("Unknown column name '%s'." % column_name)
            if column_name not in remaining_names:
                raise KGTKException("Column name '%s' was duplicated in the list." % column_name)

            if saw_ranger:
                saw_ranger = False
                prior_column_name: str = reordered_names[-1]
                prior_column_idx: int = kr.column_name_map[prior_column_name]
                column_name_idx: int = kr.column_name_map[column_name]
                start_idx: int
                end_idx: int
                idx_inc: int
                if column_name_idx > prior_column_idx:
                    start_idx = prior_column_idx + 1
                    end_idx = column_name_idx - 1
                    idx_inc = 1
                else:
                    start_idx = prior_column_idx - 1
                    end_idx = column_name_idx + 1
                    idx_inc = -1

                idx: int = start_idx
                while idx <= end_idx:
                    idx_column_name: str = kr.column_names[idx]
                    if idx_column_name not in remaining_names:
                        raise KGTKException("Column name '%s' (%s .. %s) was duplicated in the list." % (column_name, prior_column_name, column_name))
                   
                    reordered_names.append(idx_column_name)
                    remaining_names.remove(idx_column_name)
                    idx += idx_inc

            reordered_names.append(column_name)
            remaining_names.remove(column_name)

        if saw_ranger:
            raise KGTKException("The column ranger operator ('..') may not end the list of column names.")

        if len(remaining_names) > 0 and save_reordered_names is None:
            # There are remaining column names and the ellipses was not seen.
            if not omit_remaining_columns:
                raise KGTKException("No ellipses, and the following columns not accounted for: %s" % " ".join(remaining_names))
            else:
                if verbose:
                    print("Omitting the following columns: %s" % " ".join(remaining_names), file=error_file, flush=True)
        if save_reordered_names is not None:
            if len(remaining_names) > 0:
                save_reordered_names.extend(remaining_names)
            if len(reordered_names) > 0:
                save_reordered_names.extend(reordered_names)
            reordered_names = save_reordered_names

        if verbose:
            print("Opening the output file %s" % str(output_kgtk_file), file=error_file, flush=True)
        kw: KgtkWriter = KgtkWriter.open(reordered_names,
                                         output_kgtk_file,
                                         output_column_names=as_column_names,
                                         require_all_columns=True,
                                         prohibit_extra_columns=True,
                                         fill_missing_columns=False,
                                         gzip_in_parallel=False,
                                         mode=KgtkWriter.Mode[kr.mode.name],
                                         output_format=output_format,
                                         verbose=verbose,
                                         very_verbose=very_verbose,
        )

        shuffle_list: typing.List = kw.build_shuffle_list(kr.column_names)

        input_data_lines: int = 0
        row: typing.List[str]
        for row in kr:
            input_data_lines += 1
            kw.write(row, shuffle_list=shuffle_list)

        # Flush the output file so far:
        kw.flush()

        if verbose:
            print("Read %d data lines from file %s" % (input_data_lines, input_kgtk_file), file=error_file, flush=True)

        kw.close()

        return 0

    except SystemExit as e:
        raise KGTKException("Exit requested")
    except Exception as e:
        raise KGTKException(str(e))




"""
Support classes for gzip/gunzip in seperate processes.
"""

from multiprocessing import Process, Queue
import typing

from kgtk.utils.closableiter import ClosableIter

# This helper class supports running gzip in parallel.
#
# TODO: can we use attrs here?
class GzipProcess(Process):
    gzip_file: typing.TextIO

    # The line queue contains str with None as a plug.
    #
    # TODO: can we do a better job of type declaration here?
    line_queue: Queue

    GZIP_QUEUE_SIZE_DEFAULT: int = 1000

    def __init__(self,  gzip_file: typing.TextIO, line_queue: Queue):
        super().__init__()
        self.gzip_file = gzip_file
        self.line_queue = line_queue

    def run(self):
        while True:
            line: typing.Optional[str] =  self.line_queue.get()
            if line is None: # This is the plug.
                self.gzip_file.close()
                return # Exit the process.
            self.gzip_file.write(line)

    # Called from the parent process.
    def write(self, line: str):
        self.line_queue.put(line)

    # Called from the parent process.
    def close(self):
        self.line_queue.put(None) # Send the plug.
        self.join() # Wait for the plug to exit the process.

# This helper class supports running gunzip in parallel.
#
# TODO: can we use attrs here?
class GunzipProcess(Process, ClosableIter[str]):
    gzip_file: typing.TextIO

    # The line queue contains str with None as a plug.
    #
    # TODO: can we do a better job of type declaration here?
    line_queue: Queue

    GZIP_QUEUE_SIZE_DEFAULT: int = 1000

    def __init__(self,  gzip_file: typing.TextIO, line_queue: Queue):
        super().__init__()
        self.gzip_file = gzip_file
        self.line_queue = line_queue

    def run(self):
        line: str
        for line in self.gzip_file:
            self.line_queue.put(line)
        self.line_queue.put(None) # Plug the queue.

    # This is an iterator object.
    def __iter__(self)-> typing.Iterator[str]:
        return self
    
    def __next__(self)->str:
        line: typing.Optional[str] = self.line_queue.get()
        if line is None: # Have we reached the plug?
            raise StopIteration
        else:
            return line

    def close(self):
        self.gzip_file.close()




from kgtk.knowledge_graph.node import URI
from kgtk.knowledge_graph.subject import Subject
from enum import Enum
import uuid


class Rank(Enum):
    Normal = URI('wikibase:NormalRank')
    Preferred = URI('wikibase:PreferredRank')
    Deprecated = URI('wikibase:DeprecatedRank')

    # never set the rank to BestRank. This is only for truthy statements.
    # Use this only if you are sure of what you are doing
    BestRank = URI('wikibase:BestRank')


class BaseStatement(Subject):
    def __init__(self, uri):
        super().__init__(uri)

    def _add_value_node(self, prefix, p, v):
        self.add_property(URI(prefix + ':' + p), v.value)
        if v.full_value:
            self.add_property(URI(prefix + 'v:' + p), v.full_value)
        if v.normalized_value:
            self.add_property(URI(prefix + 'n:' + p), v.normalized_value)


class Statement(BaseStatement):
    def __init__(self, node_id, rank, statement_id=None):
        statement_id = statement_id or str(uuid.uuid4())
        super().__init__(URI('wds:' + node_id + '-' + statement_id))
        self.add_property(URI('rdf:type'), URI('wikibase:Statement'))
        self.add_property(URI('wikibase:rank'), rank.value)

    def add_value(self, p, v):
        self._add_value_node('ps', p, v)

    def add_qualifier(self, p, v):
        self._add_value_node('pq', p, v)

    def add_reference(self, ref):
        self.add_property(URI('prov:wasDerivedFrom'), ref)


class WDReference(BaseStatement):
    def __init__(self, reference_id=None):
        reference_id = reference_id or str(uuid.uuid4())
        super().__init__(URI('wdref:' + reference_id))
        self.add_property(URI('rdf:type'), URI('wikibase:Reference'))

    def add_value(self, p, v):
        self._add_value_node('pr', p, v)
